"""
Modeling
--------

Probabilit lets the user perform Monte-Carlo sampling using a high-level
modeling language that creates a lazy computational graph.
When a node in the graph is sampled, all ancestor nodes are sampled in turn
and samples are propagated down in the graph, from parent nodes to child nodes.

For instance, to compute the shipping cost of a box where we are uncertain
about the measurements:

>>> rng = np.random.default_rng(42)
>>> box_height = Distribution("norm", loc=0.5, scale=0.01)
>>> box_width = Distribution("norm", loc=1, scale=0.01)
>>> box_depth = Distribution("norm", loc=0.8, scale=0.01)
>>> box_volume = box_height * box_width * box_depth
>>> price_per_sqm = 50
>>> price = box_volume * price_per_sqm
>>> samples = price.sample(999, random_state=rng)
>>> float(np.mean(samples))
20.00139737515...

Distributions are built on top of scipy, so "norm" refers to the name of the
normal distribution as given in `scipy.stats`, and the arguments to the
distribution must match those given by `scipy.stats.norm` (here `loc` and
`scale`). Scipy supports almost one hundred probability distributions,
and all of them are also available in probabilit.


Composite distributions
-----------------------

Here is another example demonstrating composite distributions, where an
argument to one distribution is another distribution:

>>> eggs_per_nest = Distribution("poisson", mu=3)
>>> survivial_prob = Distribution("beta", a=10, b=15)
>>> survived = Distribution("binom", n=eggs_per_nest, p=survivial_prob)
>>> survived.sample(9, random_state=rng)
array([0., 1., 2., 0., 3., 1., 1., 0., 2.])


The modeling language
---------------------

To understand and examine the modeling language, we can perform computations
using constants. The computational graph carries out arithmetic operations
when the model is sampled. Mixing numbers with nodes in an expression is allowed,
but at least one expression or term must be a probabilit class instance:

>>> a = Constant(1)
>>> (a * 3 + 5).sample(5, random_state=rng)
array([8, 8, 8, 8, 8])
>>> Add(10, 5, 5).sample(5, random_state=rng)
array([20, 20, 20, 20, 20])

Let us build a more complicated expression with distributions:

>>> a = Distribution("norm", loc=5, scale=1)
>>> b = Distribution("expon", scale=1)
>>> expression = a**b + a * b + 5 * b

It is possible to convert distributions to scipy objects

>>> normal = Distribution("norm", loc=Constant(2)**3, scale=1)
>>> dist = normal.to_scipy()
>>> dist.rvs(5, random_state=42)
array([8.49671415, 7.8617357 , 8.64768854, 9.52302986, 7.76584663])

Every unique node in this expression can be found by calling `.nodes()`:

>>> for node in sorted(set(expression.nodes()), key=lambda n:n._id):
...     print(node)
Distribution("norm", loc=5, scale=1)
Distribution("expon", scale=1)
Power(Distribution("norm", loc=5, scale=1), Distribution("expon", scale=1))
Multiply(Distribution("norm", loc=5, scale=1), Distribution("expon", scale=1))
Add(Power(Distribution("norm", loc=5, scale=1), Distribution("expon", scale=1)), Multiply(Distribution("norm", loc=5, scale=1), Distribution("expon", scale=1)))
Constant(5)
Multiply(Distribution("expon", scale=1), Constant(5))
Add(Add(Power(Distribution("norm", loc=5, scale=1), Distribution("expon", scale=1)), Multiply(Distribution("norm", loc=5, scale=1), Distribution("expon", scale=1))), Multiply(Distribution("expon", scale=1), Constant(5)))

Sampling any node is done by calling the `.sample()` method:

>>> expression.sample(5, random_state=rng)
array([ 2.70764145, 36.58578812,  7.07064239,  1.84433247,  3.90951632])

Sampling the expression has the side effect that `.samples_` is populated on
*every* ancestor node in the expression, for instance:

>>> a.samples_
array([4.51589278, 4.37788659, 5.25960812, 5.80609507, 4.33770499])

Here is an even more complex expression, showcasing some mathematical functions:

>>> a = Distribution("norm", loc=0, scale=1)
>>> b = Distribution("norm", loc=0, scale=2)
>>> c = Distribution("norm", loc=0, scale=3)
>>> expression = a*a - Add(a, b, c) + Abs(b)**Abs(c) + Exp(1 / Abs(c))
>>> expression.sample(5, random_state=rng)
array([ 4.70542018, 14.43250192,  6.74494838, -0.14020459, -3.27334554])

Nodes are hashable and can be used in sets, so __hash__ and __eq__ must both
be defined. Therefore we cannot use `==` for modeling; equality in that context
has another meaning. Use the Equal node instead. This is only relevant in cases
when equality operators is part of a model. For distribution defined on real
numbers (e.g. Normal) equality does not make sense since the probability that two
floats are equal is zero.

>>> dice1 = Distribution("uniform", loc=1, scale=6) // 1
>>> dice2 = Distribution("uniform", loc=1, scale=6) // 1
>>> equal_result = Equal(dice1, dice2)
>>> float(equal_result.sample(999, random_state=42).mean())
0.166...

Empirical distributions may also be used. They wrap np.quantile and take the
same arguments. For instance, to sample a dice we can pass `closest_observation`:

>>> dice = EmpiricalDistribution([1, 2, 3, 4, 5, 6], method="closest_observation")
>>> dice.sample(9, random_state=42)
array([2, 6, 4, 4, 1, 1, 1, 5, 4])

To sample from a non-parametric distribution defined by data,
similar to a kernel density estimate:

>>> cost = EmpiricalDistribution([200, 200, 300, 250, 225])
>>> cost.sample(9, random_state=42)
array([212.45401188, 290.14286128, 248.19939418, 234.86584842,
       200.        , 200.        , 200.        , 273.23522915,
       235.11150117])


Correlations
------------
To correlate samples, use the `.correlate()` method on any node that is a
descendant of the nodes you wish to correlate:

>>> from scipy.stats import pearsonr
>>> a, b = Distribution("uniform"), Distribution("expon")
>>> corr_mat = np.array([[1, 0.5], [0.5, 1]])
>>> s = (a + b).correlate(a, b, corr_mat=corr_mat).sample(99, random_state=0)

We can verify that the correlation is close to the desired value of 0.5:

>>> float(pearsonr(a.samples_, b.samples_).statistic)
0.49245...


Multivariate distributions
--------------------------
Support for multivariate distributions (MVD) is implemented, but is limited:

  1. the MVD must be a source node (its arguments cannot be other distributions)
  2. its return values *must* be unpacked as marginals (slices)
  3. only pseudo-random sampling is possible (LHS, Sobol, etc. is ignored)

For instance, to create a Dirichlet distribution, we must unpack it as follows:

>>> d1, d2 = MultivariateDistribution("dirichlet", alpha=[1, 2])
>>> d1
MarginalDistribution(Distribution("dirichlet", alpha=[1, 2]), d=0)

Since the Direchlet distribution is defined on an (n-1) dimensional simplex,
the sum of the marginals is always 1. We can check this by computing:

>>> (d1 + d2).sample(5, random_state=0)
array([1., 1., 1., 1., 1.])

Each marginal has values between 0 and 1:

>>> d2.samples_.round(3)
array([0.559, 0.324, 0.284, 0.524, 0.599])

Here is an example with a multivariate normal distribution:

>>> cov = np.array([[1, 0.5], [0.5, 1]])
>>> n1, n2 = MultivariateDistribution("multivariate_normal", mean=[1, 2], cov=cov)
>>> n1.sample(5, random_state=0)
array([0.72058767, 3.13703525, 2.38930155, 1.50866787, 0.77018653])
>>> (n1 + n2).sample(5, random_state=0)
array([2.52848604, 5.31650094, 5.20076878, 4.06217341, 1.40748585])


Samplers
--------

By default sampling uses pseudo-random numbers. To use e.g. latin hybercube
sampling, pass the `method` argument into `.sample()`.

>>> dice = EmpiricalDistribution([1, 2, 3, 4, 5, 6], method="closest_observation")
>>> float(dice.sample(9, random_state=1, method="lhs").mean())
3.222...
>>> float(dice.sample(9, random_state=1, method=None).mean())
1.888...

To retain more control, use the `sample_from_quantiles` method directly instead.
The quantiles are passed to the inverse CDF (percent point function) when sampling:

>>> from scipy.stats.qmc import LatinHypercube
>>> d = expression.num_distribution_nodes()
>>> hypercube = LatinHypercube(d=d, rng=rng, optimization="random-cd")
>>> hypercube_samples = hypercube.random(5) # Draw 5 samples
>>> expression.sample_from_quantiles(hypercube_samples)
array([ 7.80785741,  3.72416016,  3.77849849, -3.83561905, 38.02479019])


Garbage collection
------------------
By default the `.samples_` attribute is set on every ancestor when sampling.
This can lead to memory overhead in large graphs. A garbage collection
strategy can be set to keep the memory constrained:

Setting `gc_strategy=[]` removes `.samples_` on all nodes except the final node:

>>> a = Distribution("norm")
>>> intermediate_result = (a + a)**2 - a
>>> final_result = Exp(intermediate_result)
>>> final_result.sample(3, random_state=42, gc_strategy=[]).round(3)
array([2.0730000e+00, 1.0532374e+04, 2.4920000e+00])

Verif that `.samples_` was not set on ancestors:

>>> hasattr(a, "samples_")
False
>>> hasattr(intermediate_result, "samples_")
False

To keep intermediate results on some nodes, pass them into `gc_strategy`:

>>> final_result.sample(3, random_state=42, gc_strategy=[a]).round(3)
array([2.0730000e+00, 1.0532374e+04, 2.4920000e+00])
>>> hasattr(a, "samples_")
True
>>> hasattr(intermediate_result, "samples_")
False

When passing `gc_strategy=None` (the default), probabilit will not do any GC.


Functions
---------

If you have a function that is not an arithmetic expression, you can still
Monte-Carlo simulate through it with the `scalar_transform` decorator, which
will pass each sample through the computation node in a loop:

>>> def function(a, b):
...     if a > 0:
...         return a * b
...     else:
...         return 0
>>> function = scalar_transform(function)

Now we can create a computational graph:

>>> a = Distribution("norm", loc=0, scale=1)
>>> b = Distribution("norm", loc=0, scale=2)
>>> expression = function(a, b) # Function is not actually evaluated here

Now sample 'through' the function:

>>> expression.sample(5, random_state=rng)
array([0.        , 0.        , 0.45555522, 0.        , 0.        ])

This type of sampling is slower, since it passes each sample in turn and
is not vectorized. The advantage is that the function can be a complex
simulation model, or read and write files, etc.

"""

import operator
import functools
import numpy as np
import scipy as sp
import numbers
from scipy import stats
import abc
import itertools
import networkx as nx
from scipy._lib._util import check_random_state
from probabilit.correlation import (
    nearest_correlation_matrix,
    ImanConover,
    Cholesky,
    Composite,
    Permutation,
)
from probabilit.utils import build_corrmat, zip_args
from probabilit.garbage_collector import GarbageCollector
import copy


# =============================================================================
# FUNCTIONS
# =============================================================================


def python_to_prob(argument):
    """Convert basic Python types to probabilit types."""
    if isinstance(argument, numbers.Number):
        return Constant(argument)
    elif isinstance(argument, Node):
        return argument
    else:
        raise ValueError(f"Type not compatible with probabilit: {argument}")


# =============================================================================
# COMPUTATIONAL GRAPH AND MODELING LANGUAGE
# =============================================================================
#
# There are three main types of Node instances, they are:
#   - Constant:      numbers like 2 or 5.5, which are always source nodes in the graph
#   - Distribution:  typically source nodes, but can have parents if composite
#   - Transform:     arithmetic operations like + or **, or general functions,
#                    must always have parents
#
# An expression such as:
#
#  mu = Distribution("norm", loc=0, scale=1)
#  normal  = Distribution("norm", loc=mu, scale=1)
#  result = mu + normal - 2
#
# Is represented by a graph such as:
#
#        mu --> normal
#         \       /
#          \     /
#           v   v
#             +         2
#              \       /
#               \     /
#                v   v
#                  -
#               (result)
#
# Where:
#   * "mu" is a Distribution and a source node
#   * "normal" is a Distribution, but not a source node (has "mu" as parent)
#   * "+" is a Transform (has "mu" and "normal" as parents)
#   * "2" is a Constant and a source node
#   * "-" (the result) is a Transform (has "+" and "2" as parents)
#
# Some further terminology:
#   * The _ancestors_ of node "-" are {"+", "2", "mu", "normal"}
#   * A node is said to be an _initial sampling node_ iff
#      (1) The node is a Distribution
#      (2) None of its ancestors are Distributions
#     For instance, in the graph above, the node "mu" is an initial sampling node.
#     Initial sampling nodes are the nodes that we can impose correlations on.
#     We cannot impose correlations on "normal" above, since its correlation
#     is determined by the graph structure.
#   * result.sample() samples the expression "mu + normal - 2" by propagating
#     through the graph, parents first. More specifically, each category of nodes
#     (Constant, Transform and Distribution) have their own internal _sample methods
#     and we propagate the sampling of the graph in the following way:
#      (1) "mu" is sampled first by calling the internal _sample method
#          of the Distribution class.
#      (2) Then "normal" is sampled using the samples of "mu" and the same
#          _sample method as in (1).
#      (3) Thereafter, "+" is sampled using the samples of both "mu" and "normal",
#          and the internal _sample method of the Transform class.
#      (4) Finally, "-" is sampled last, using the samples of "+" and "2",
#          and the same _sample method as in (3).
#     Since the node "2" has no parents (as is the case for all Constant nodes),
#     it can be sampled at any time prior to sampling "-".
#     A topological ordering of the graph makes sure that parents are always
#     sampled first, and the internal _sample methods of the Distributon and
#     the Transform class rely on the samples of parent nodes (in the case where
#     parents exist).


class Node(abc.ABC):
    """A node in the computational graph."""

    id_iter = itertools.count()  # Every node gets a unique ID

    def __init__(self):
        self._id = next(self.id_iter)
        self._correlations = []

    def __eq__(self, other):
        if not isinstance(other, Node):
            return NotImplemented
        # Needed for set() to work on Node. Equality in models must use Equal()
        return self._id == other._id

    def __hash__(self):
        return self._id

    def copy(self):
        """Copy the Node, including the entire graph above it.

        Examples
        --------
        >>> mu = Distribution("norm", loc=0, scale=1)
        >>> a = Distribution("norm", loc=mu, scale=Constant(0.5))
        >>> a2 = a.copy()
        >>> a is a2
        False
        >>> a2.kwargs["loc"] == a.kwargs["loc"]
        True
        >>> a2.kwargs["loc"] is a.kwargs["loc"]
        False
        """
        # Map from ID to new object copy
        id_to_new = dict()

        def update(item):
            """Given an item, use the ID to map to new object copy."""
            if isinstance(item, Node):
                return id_to_new[item._id]
            return copy.deepcopy(item)

        # Go through nodes in topologial order, guaranteeing that parents
        # have always been copied over to new graph when children are copied.
        for node in nx.topological_sort(self.to_graph()):
            # Copy the node itself and update the mapping
            copied = copy.copy(node)  # Copy a node WITHOUT copying the graph
            id_to_new[copied._id] = copied

            # Copy samples if they exist
            if hasattr(copied, "samples_"):
                copied.samples_ = np.copy(copied.samples_)

            copied._correlations = copy.deepcopy(copied._correlations)

            # Now that the node has been updated, update references to parents
            # to point to Nodes in the new copied graph instead of the old one.
            if isinstance(copied, (AbstractDistribution, ScalarFunctionTransform)):
                copied.args = tuple(update(arg) for arg in copied.args)
                copied.kwargs = {k: update(v) for (k, v) in copied.kwargs.items()}
            elif isinstance(copied, (VariadicTransform, BinaryTransform)):
                copied.parents = tuple(update(p) for p in copied.parents)
            elif isinstance(copied, UnaryTransform):
                copied.parent = update(copied.parent)
            elif isinstance(copied, Constant):
                copied.value = update(copied.value)
            else:
                pass

        return id_to_new[self._id]

    def nodes(self):
        """Yields `self` and all ancestors using depth-first-search.

        Examples
        --------
        >>> expression = Distribution("norm") -  2**Constant(2)
        >>> for node in expression.nodes():
        ...     print(node)
        Subtract(Distribution("norm"), Power(Constant(2), Constant(2)))
        Power(Constant(2), Constant(2))
        Constant(2)
        Constant(2)
        Distribution("norm")
        """
        queue = [(self)]
        while queue:
            yield (node := queue.pop())
            queue.extend(node.get_parents())

    def num_distribution_nodes(self):
        """Number of unique ancestor nodes that are distribution nodes."""
        return sum(
            1 for node in set(self.nodes()) if isinstance(node, AbstractDistribution)
        )

    def sample(
        self,
        size=1,
        *,
        random_state=None,
        method=None,
        correlator="composite",
        gc_strategy=None,
    ):
        """Sample the current node and assign attribute `samples_` to nodes.

        Parameters
        ----------
        size : int, optional
            Number of samples to draw.
        random_state : np.random.Generator, int or None, optional
            A random state for the random number generator. The default is None.
        method : str, optional
            Sampling method, one of "lhs" (qmc.LatinHypercube), "halton"
            (qmc.Halton) or "sobol" (qmc.Sobol). The default is None, which
            means pseudo-random sampling.
        correlator : Correlator or str, optional
            A Correlator instance or a string in {"cholesky", "imanconover",
           "permutation", "composite"}. The default is "composite", which first
            runs Iman-Conover, then runs the Permutation correlator on the result.
        gc_strategy : None or list, optional
            If None, no garbage collection is performed and the attribute
            `.samples_` will be set on all nodes. If an empty list [], then
            all nodes except the final one will be garbage collected. If a list
            of Node instances, then those will be garbage collected.

        Returns
        -------
        np.ndarray
            An array of samples, with length `size`.

        Examples
        --------
        >>> result = 2 * Distribution("expon", scale=1/3)
        >>> result.sample(random_state=0)
        array([0.53058301])
        >>> result.sample(size=5, random_state=0)
        array([0.53058301, 0.83728718, 0.6154821 , 0.52480077, 0.36736566])
        >>> result.sample(size=5, random_state=0, method="lhs")
        array([1.11212876, 0.273718  , 0.03808862, 0.5702549 , 0.83779147])

        Set a custom correlator by giving a Correlator type.
        The API of a correlator is:

            1. correlator = Correlator(correlation_matrix)
            2. X_corr = correlator(X_samples)  # shape (samples, variable)

        >>> from probabilit.correlation import Cholesky, ImanConover
        >>> from scipy.stats import pearsonr
        >>> a, b = Distribution("uniform"), Distribution("expon")
        >>> corr_mat = np.array([[1, 0.6], [0.6, 1]])
        >>> result = (a + b).correlate(a, b, corr_mat=corr_mat)

        >>> s = result.sample(25, random_state=0, correlator=Cholesky())
        >>> float(pearsonr(a.samples_, b.samples_).statistic)
        0.600000...
        >>> float(np.min(b.samples_)) # Cholesky does not preserve marginals!
        -0.35283...

        >>> s = result.sample(25, random_state=0, correlator=ImanConover())
        >>> float(pearsonr(a.samples_, b.samples_).statistic)
        0.617109...
        >>> float(np.min(b.samples_)) # ImanConover does preserve marginals
        0.062115...
        """
        if not isinstance(size, numbers.Integral):
            raise TypeError("`size` must be a positive integer")
        if not size > 0:
            raise ValueError("`size` must be a positive integer")

        d = self.num_distribution_nodes()  # Dimensionality of sampling

        # Draw a quantiles of random variables in [0, 1] using a method
        methods = {
            "lhs": sp.stats.qmc.LatinHypercube,
            "halton": sp.stats.qmc.Halton,
            "sobol": sp.stats.qmc.Sobol,
        }
        if method is None:  # Pseudo-random sampling
            random_state = check_random_state(random_state)
            quantiles = random_state.random((size, d))
        else:  # Quasi-random sampling
            sampler = methods[method.lower().strip()](d=d, rng=random_state)
            quantiles = sampler.random(n=size)

        return self.sample_from_quantiles(
            quantiles,
            correlator=correlator,
            gc_strategy=gc_strategy,
            random_state=random_state,
        )

    def sample_from_quantiles(
        self, quantiles, *, correlator="composite", gc_strategy=None, random_state=None
    ):
        """Use samples from an array of quantiles in [0, 1] to sample all
        distributions. The array must have shape (dimensionality, num_samples)."""
        assert nx.is_directed_acyclic_graph(self.to_graph())
        size, n_dim = quantiles.shape
        assert n_dim == self.num_distribution_nodes()

        # Get the correct correlator class based on strings
        CORRELATOR_MAP = {
            "imanconover": ImanConover(),
            "cholesky": Cholesky(),
            "permutation": Permutation(random_state=random_state),
            "composite": Composite(random_state=random_state),
        }
        if isinstance(correlator, str):
            correlator = correlator.lower().strip()
            if correlator not in CORRELATOR_MAP.keys():
                raise ValueError(
                    f"`{correlator=}` must be in {set(CORRELATOR_MAP.keys())}"
                )
            correlator = CORRELATOR_MAP[correlator]  # Map to instance

        # Prepare columns of quantiles, one column for each Distribution
        columns = iter(list(quantiles.T))

        # Clear any samples that might exist in the graph
        for node in set(self.nodes()):
            if hasattr(node, "samples_"):
                delattr(node, "samples_")

        # Set up garbage collection
        gc = GarbageCollector(strategy=gc_strategy).set_sink(self)

        # Start with initial sampling nodes, which contain independent variables
        initial_sampling_nodes = set(
            node for node in self.nodes() if node._is_initial_sampling_node()
        )
        # Ensure consistent ordering for reproducible results
        initial_sampling_nodes = sorted(initial_sampling_nodes, key=lambda n: n._id)

        # Loop over all initial sampling nodes (ISN) and sample them
        G = self.to_graph()
        for node in initial_sampling_nodes:
            # Sample all ancestors of ISNs
            ancestors = G.subgraph(nx.ancestors(G, node))
            for ancestor in nx.topological_sort(ancestors):
                assert isinstance(ancestor, (Constant, Transform))
                ancestor.samples_ = ancestor._sample(size=size)

            # Sample the ISN
            assert isinstance(node, AbstractDistribution)
            node.samples_ = node._sample(q=next(columns))

        # Go through all ancestor nodes and create a list [(var, corr), ...]
        # that contains all correlations we must induce
        correlations = []
        for node in set(self.nodes()):
            if hasattr(node, "_correlations"):
                correlations.extend(node._correlations)

        # Check that each variable to correlate is an initial sampling node
        for variables, _ in correlations:
            for variable in variables:
                if variable not in initial_sampling_nodes:
                    raise ValueError(f"Cannot correlate variable: {variable}")

        # Check that no correlation has been specified twice
        variable_sets = [set(variables) for (variables, _) in correlations]
        for vars1, vars2 in itertools.combinations(variable_sets, 2):
            common = vars1.intersection(vars2)
            if len(common) > 1:
                raise ValueError(f"Correlations specified more than once: {common}")

        # Map all variables to integers to associate them with a column
        all_variables = list(functools.reduce(set.union, variable_sets, set()))
        # Ensure consistent ordering for reproducible results
        all_variables = sorted(all_variables, key=lambda n: n._id)
        var_to_int = {v: i for (i, v) in enumerate(all_variables)}
        correlations = [
            (tuple(var_to_int[var] for var in variables), corrmat)
            for (variables, corrmat) in correlations
        ]

        # If there are any correlations to induce, do so
        if correlations:
            # Induce correlations
            correlation_matrix = build_corrmat(correlations)
            correlation_matrix = nearest_correlation_matrix(correlation_matrix)

            # Set the target (goal) correlation matrix
            correlator = correlator.set_target(correlation_matrix)

            # Concatenate samples, correlate them (shift rows in each col), then re-assign
            samples_input = np.vstack([var.samples_ for var in all_variables]).T
            samples_ouput = correlator(samples_input)
            for var, sample in zip(all_variables, samples_ouput.T):
                var.samples_ = np.copy(sample)

        # Iterate sampling though the graph
        for node in nx.topological_sort(G):
            if hasattr(node, "samples_"):  # Skip if samples already exists
                pass
            elif isinstance(node, Constant):
                node.samples_ = node._sample(size=size)  # Draw constants
            elif isinstance(node, AbstractDistribution):
                node.samples_ = node._sample(q=next(columns))  # Sample distr
            elif isinstance(node, Transform):
                node.samples_ = node._sample()  # Propagate through transform
            else:
                raise TypeError(
                    "Node must be Constant, AbstractDistribution or Transform."
                )

            is_numeric = (node.samples_ is not None) and np.issubdtype(
                node.samples_.dtype, np.number
            )
            if is_numeric and not np.all(np.isfinite(node.samples_)):
                raise ValueError(
                    f"Sampling this node gave non-finite values: {node}\n{node.samples_}"
                )

            # Tell the garbage collector that we sampled this node.
            # If the reference counter reaches zero (a parent has no unsampled
            # children), then the `.samples_` attribute of the parent might
            # be deleted (only if the garbage collection strategy allows it).
            gc.decrement_and_delete(node)

        return self.samples_

    def _is_initial_sampling_node(self):
        """A node is an initial sample node iff:
        (1) It is a Distribution
        (2) None of its ancestors are Distributions (all are Constant/Transform)"""

        is_distribution = isinstance(self, AbstractDistribution)
        ancestors = set(self.nodes()) - set([self])
        ancestors_distr = any(
            isinstance(node, AbstractDistribution) for node in ancestors
        )
        return is_distribution and not ancestors_distr

    def correlate(self, *variables, corr_mat):
        """Store correlations on variables.

        When `.correlate(*variables)` is called on a node, the variables must
        be ancestors of that node. The order of the variables should match the
        order of the rows/columns in the correlation matrix.

        Examples
        --------
        >>> a = Distribution("expon", 1)
        >>> b = Distribution("poisson", 1)
        >>> corr_mat = np.array([[1, 0.5], [0.5, 1]])

        Correlations can be induced on any child node.

        >>> import scipy as sp
        >>> result = (a + b)
        >>> result = result.correlate(a, b, corr_mat=corr_mat)
        >>> _ = result.sample(999, random_state=0)
        >>> float(sp.stats.pearsonr(a.samples_, b.samples_).statistic)
        0.483035...

        """
        assert corr_mat.ndim == 2
        assert corr_mat.shape[0] == corr_mat.shape[1]
        assert corr_mat.shape[0] == len(variables)
        assert len(variables) == len(set(variables))
        nodes = set(self.nodes())
        for var in variables:
            if var not in nodes:
                raise ValueError(f"{var} is not an ancestor of {self}")

        valid_diag = np.allclose(np.diag(corr_mat), 1)
        valid_entries = np.allclose(np.clip(corr_mat, -1, 1), corr_mat)
        if not (valid_diag and valid_entries):
            raise ValueError(
                "Correlation matrix must have entries in [-1, 1] and 1 on diagonal."
            )

        self._correlations.append((list(variables), np.copy(corr_mat)))
        return self

    def to_graph(self):
        """Convert the computational graph to a networkx MultiDiGraph."""
        nodes = list(self.nodes())

        # Special case if there is only one node
        if len(nodes) == 1:
            G = nx.MultiDiGraph()
            G.add_node(self)
            return G

        # General case
        edge_list = [
            (ancestor, node)
            for node in nodes
            for ancestor in node.get_parents()
            if not node.is_source_node
        ]
        return nx.MultiDiGraph(edge_list)


class OverloadMixin:
    """Overloads dunder (double underscore) methods for easier modeling."""

    def __add__(self, other):
        return Add(self, other)

    def __radd__(self, other):
        return Add(self, other)

    def __mul__(self, other):
        return Multiply(self, other)

    def __rmul__(self, other):
        return Multiply(self, other)

    def __floordiv__(self, other):
        return FloorDivide(self, other)

    def __rfloordiv__(self, other):
        return FloorDivide(other, self)

    def __truediv__(self, other):
        return Divide(self, other)

    def __rtruediv__(self, other):
        return Divide(other, self)

    def __mod__(self, other):
        return Mod(self, other)

    def __rmod__(self, other):
        return Mod(other, self)

    def __sub__(self, other):
        return Subtract(self, other)

    def __rsub__(self, other):
        return Subtract(other, self)

    def __pow__(self, other):
        return Power(self, other)

    def __rpow__(self, other):
        return Power(other, self)

    def __neg__(self):
        return Negate(self)

    def __abs__(self):
        return Abs(self)

    def __lt__(self, other):
        return LessThan(self, other)

    def __le__(self, other):
        return LessThanOrEqual(self, other)

    def __gt__(self, other):
        return GreaterThan(self, other)

    def __ge__(self, other):
        return GreaterThanOrEqual(self, other)

    # TODO: __eq__ (==) and __ne__ (!=) are not implemented here,
    # because they are used in set(nodes), which relies upon
    # both equality checks and __hash__.


class Constant(Node, OverloadMixin):
    """A constant is a number or a string. If the value is a string, sampling returns an array of the string value.

    Examples
    --------
    >>> Constant(2)._sample(5)
    array([2, 2, 2, 2, 2])
    >>> Constant("car")._sample(5)
    array(['car', 'car', 'car', 'car', 'car'], dtype='<U3')
    """

    is_source_node = True  # A Constant is always a source node

    def __init__(self, value):
        self.value = value.value if isinstance(value, Constant) else value
        super().__init__()

    def _sample(self, size=None):
        if size is None:
            return self.value
        return np.array([self.value] * size)

    def get_parents(self):
        yield from []  # A Constant does not have any parents

    def __repr__(self):
        return f"{type(self).__name__}({self.value})"


class AbstractDistribution(Node, OverloadMixin, abc.ABC):
    pass


class Distribution(AbstractDistribution):
    """A distribution is a sampling node with or without ancestors."""

    def __init__(self, distr, *args, **kwargs):
        self.distr = distr
        self.args = args
        self.kwargs = kwargs
        super().__init__()

    def __repr__(self):
        args = ", ".join(repr(arg) for arg in self.args)
        kwargs = ", ".join(f"{k}={repr(v)}" for (k, v) in self.kwargs.items())
        out = f'{type(self).__name__}("{self.distr}"'
        if args:
            out += f", {args}"
        if kwargs:
            out += f", {kwargs}"
        return out + ")"

    def to_scipy(self):
        if not self._is_initial_sampling_node():
            raise Exception(
                "To convert a distribution to a scipy object, "
                "it must be an initial sampling node (no ancestors can be Distributions)"
            )

        node = self.copy()  # do not mutate self

        try:
            distribution = getattr(stats, node.distr)
        except AttributeError:
            raise AttributeError(f"{self.distr!r} is not a valid scipy distribution")

        def to_number(arg):
            """Unpack argument to a number in case parents are Constant/Transform"""
            return arg.sample(1)[0] if isinstance(arg, Node) else arg

        args = tuple(to_number(arg) for arg in node.args)
        kwargs = {k: to_number(v) for (k, v) in node.kwargs.items()}

        return distribution(*args, **kwargs)

    def _sample(self, q):
        def unpack(arg):
            """Unpack distribution arguments (parents) to arrays if Node."""
            return arg.samples_ if isinstance(arg, Node) else arg

        # Parse the arguments and keyword arguments for the distribution
        args = tuple(unpack(arg) for arg in self.args)
        kwargs = {k: unpack(v) for (k, v) in self.kwargs.items()}

        # Sample from the distribution with inverse CDF
        distribution = getattr(stats, self.distr)
        try:
            return distribution(*args, **kwargs).ppf(q)
        except AttributeError:
            # Multivariate distributions do not have .ppf()
            # isinstance(distribution, (multi_rv_generic, multi_rv_frozen))
            seed = int(q[0] * 2**20)  # Seed based on q
            return distribution(*args, **kwargs).rvs(size=len(q), random_state=seed)

    def get_parents(self):
        # A distribution only has parents if it has parameters that are Nodes
        for arg in self.args + tuple(self.kwargs.values()):
            if isinstance(arg, Node):
                yield arg

    @property
    def is_source_node(self):
        return list(self.get_parents()) == []


class EmpiricalDistribution(AbstractDistribution):
    """A distribution is a sampling node with or without ancestors.

    A thin wrapper around numpy.quantile."""

    is_source_node = True

    def __init__(self, data, **kwargs):
        self.data = np.array(data)
        self.kwargs = kwargs
        super().__init__()

    def __repr__(self):
        return f"{type(self).__name__}()"

    def _sample(self, q):
        return np.quantile(a=self.data, q=q, **self.kwargs)

    def get_parents(self):
        yield from []  # A EmpiricalDistribution does not have any parents


class CumulativeDistribution(AbstractDistribution):
    """A distribution defined by cumulative quantiles.

    Examples
    --------
    >>> distr = CumulativeDistribution([0, 0.2, 0.8, 1], [10, 15, 20, 25])
    >>> distr._sample(np.linspace(0, 1, num=6))
    array([10.        , 15.        , 16.66666667, 18.33333333, 20.        ,
           25.        ])
    >>> distr.sample(9, random_state=42)
    array([16.45450099, 23.76785766, 19.43328285, 18.32215403, 13.90046601,
           13.89986301, 11.4520903 , 21.65440364, 18.3426251 ])
    """

    is_source_node = True

    def __init__(self, quantiles, cumulatives):
        self.q = np.array(quantiles)
        self.cumulatives = np.array(cumulatives)
        if not np.all(np.diff(self.q) > 0):
            raise ValueError("The quantiles must be strictly increasing.")
        if not np.all(np.diff(self.cumulatives) > 0):
            raise ValueError("The cumulatives must be strictly increasing.")
        if not (np.isclose(np.min(self.q), 0) and np.isclose(np.max(self.q), 1)):
            raise ValueError("Lowest quantile must be 0 and highest must be 1.")
        super().__init__()

    def __repr__(self):
        return f"{type(self).__name__}(quantiles={repr(self.q)}, cumulatives={repr(self.cumulatives)})"

    def _sample(self, q):
        # Inverse CDF sampling
        return np.interp(x=q, xp=self.q, fp=self.cumulatives)

    def get_parents(self):
        yield from []


class DiscreteDistribution(AbstractDistribution):
    """A discrete (or categorical) distribution defined by values and probabilities.

    Examples
    --------
    >>> distr = DiscreteDistribution([10, 15, 20], probabilities=[0.2, 0.3, 0.5])
    >>> distr._sample(np.linspace(0, 1, num=5, endpoint=False))
    array([10, 15, 15, 20, 20])
    >>> distr = DiscreteDistribution(["A", "B", "C", "D", "E", "F"])
    >>> distr.sample(9, random_state=42)
    array(['C', 'F', 'E', 'D', 'A', 'A', 'A', 'F', 'D'], dtype='<U1')
    """

    is_source_node = True

    def __init__(self, values, probabilities=None):
        self.values = np.array(values)
        if probabilities is None:
            self.probabilities = np.ones(len(self.values), dtype=float)
            self.probabilities = self.probabilities / np.sum(self.probabilities)
        else:
            self.probabilities = np.array(probabilities)

        if not len(self.values) == len(self.probabilities):
            raise ValueError(
                f"Length mismatch: {len(self.values)=}  {len(self.probabilities)=}"
            )
        if not np.isclose(np.sum(self.probabilities), 1.0):
            raise ValueError(f"Probabilities must sum to 1. {sum(self.probabilities)=}")
        if np.any(self.probabilities < 0):
            raise ValueError("Probabilities are not non-negative.")
        super().__init__()

    def __repr__(self):
        return f"{type(self).__name__}(values={repr(self.values)}, probabilities={repr(self.probabilities)})"

    def _sample(self, q):
        cumulative_probabilities = np.cumsum(self.probabilities)
        idx = np.searchsorted(cumulative_probabilities, v=q, side="right")
        return self.values[idx]

    def get_parents(self):
        yield from []


# ========================================================


class Transform(Node, OverloadMixin, abc.ABC):
    """Transform nodes represent arithmetic operations."""

    is_source_node = False

    def __repr__(self):
        parents = ", ".join(repr(parent) for parent in self.get_parents())
        return f"{type(self).__name__}({parents})"


class VariadicTransform(Transform):
    """Parent class for variadic transforms (must be associative), e.g.
    Add(arg1, arg2, arg3, arg4, ...)
    Multiply(arg1, arg2, arg3, arg4, ...)

    """

    def __init__(self, *args):
        self.parents = tuple(python_to_prob(arg) for arg in args)
        super().__init__()

    def _sample(self, size=None):
        samples = (parent.samples_ for parent in self.parents)
        return functools.reduce(self.op, samples)

    def get_parents(self):
        yield from self.parents


class Add(VariadicTransform):
    op = operator.add


class Multiply(VariadicTransform):
    op = operator.mul


class Max(VariadicTransform):
    op = np.maximum


class Min(VariadicTransform):
    op = np.minimum


class All(VariadicTransform):
    op = np.logical_and


class Any(VariadicTransform):
    op = np.logical_or


class Avg(VariadicTransform):
    def _sample(self, size=None):
        # Avg(a, Avg(b, c)) !=  Avg(Avg(a, b), c), so we override _sample()
        samples = tuple(parent.samples_ for parent in self.parents)
        return np.average(np.vstack(samples), axis=0)


class NoOp(VariadicTransform):
    """Sample all ancestor variables, but do nothing else."""

    def _sample(self, size=None):
        tuple(parent.samples_ for parent in self.parents)


class BinaryTransform(Transform):
    """Class for binary transforms, such as Divide, Power, Subtract, etc."""

    def __init__(self, *args):
        self.parents = tuple(python_to_prob(arg) for arg in args)
        super().__init__()

    def _sample(self, size=None):
        samples = (parent.samples_ for parent in self.parents)
        return self.op(*samples)

    def get_parents(self):
        yield from self.parents


class FloorDivide(BinaryTransform):
    op = np.floor_divide


class Mod(BinaryTransform):
    op = np.mod


class Divide(BinaryTransform):
    op = operator.truediv


class Power(BinaryTransform):
    op = operator.pow


class Subtract(BinaryTransform):
    op = operator.sub


class Equal(BinaryTransform):
    op = np.equal


class NotEqual(BinaryTransform):
    op = np.not_equal


class LessThan(BinaryTransform):
    op = operator.lt


class LessThanOrEqual(BinaryTransform):
    op = operator.le


class GreaterThan(BinaryTransform):
    op = operator.gt


class GreaterThanOrEqual(BinaryTransform):
    op = operator.ge


class IsClose(BinaryTransform):
    op = np.isclose


class UnaryTransform(Transform):
    """Class for unary tranforms, i.e. functions that take one argument, such
    as Abs(), Exp(), Log()."""

    def __init__(self, arg):
        self.parent = python_to_prob(arg)
        super().__init__()

    def _sample(self, size=None):
        return self.op(self.parent.samples_)

    def get_parents(self):
        yield self.parent


class Negate(UnaryTransform):
    op = operator.neg


class Abs(UnaryTransform):
    op = operator.abs


class Log(UnaryTransform):
    op = np.log


class Exp(UnaryTransform):
    op = np.exp


class Floor(UnaryTransform):
    op = np.floor


class Ceil(UnaryTransform):
    op = np.ceil


class Sign(UnaryTransform):
    op = np.sign


class Sqrt(UnaryTransform):
    op = np.sqrt


class Square(UnaryTransform):
    op = np.square


class Log10(UnaryTransform):
    op = np.log10


# Trigonometric functions
class Sin(UnaryTransform):
    op = np.sin


class Cos(UnaryTransform):
    op = np.cos


class Tan(UnaryTransform):
    op = np.tan


class Arcsin(UnaryTransform):
    op = np.arcsin


class Arccos(UnaryTransform):
    op = np.arccos


class Arctan(UnaryTransform):
    op = np.arctan


class Arctan2(BinaryTransform):
    op = np.arctan2


# Hyperbolic functions
class Sinh(UnaryTransform):
    op = np.sinh


class Cosh(UnaryTransform):
    op = np.cosh


class Tanh(UnaryTransform):
    op = np.tanh


class Arcsinh(UnaryTransform):
    op = np.arcsinh


class Arccosh(UnaryTransform):
    op = np.arccosh


class Arctanh(UnaryTransform):
    op = np.arctanh


class ScalarFunctionTransform(Transform):
    """A general-purpose transform using a function that takes scalar arguments
    and returns a scalar result."""

    def __init__(self, func, args, kwargs):
        self.func = func
        self.args = args
        self.kwargs = kwargs
        super().__init__()

    def _sample(self, size=None):
        def unpack(arg):
            return arg.samples_ if isinstance(arg, Node) else itertools.repeat(arg)

        # Sample arguments
        args = tuple(unpack(arg) for arg in self.args)
        kwargs = {k: unpack(v) for (k, v) in self.kwargs.items()}

        return np.array(
            [
                self.func(*args_i, **kwargs_i)
                for (args_i, kwargs_i) in zip_args(args, kwargs)
            ]
        )

    def get_parents(self):
        # A function has have parents if its arguments are Nodes
        for arg in self.args + tuple(self.kwargs.values()):
            if isinstance(arg, Node):
                yield arg


def scalar_transform(func):
    """Transform a function, so that when it is called it is converted to
    a ScalarFunctionTransform."""

    @functools.wraps(func)
    def transformed_function(*args, **kwargs):
        return ScalarFunctionTransform(func, args, kwargs)

    return transformed_function


class MarginalDistribution(Transform):
    """A maginal distribution is a 'slice' of a multivariate distribution.

    Examples
    --------
    >>> distr = Distribution("multinomial", n=10, p=[0.1, 0.2, 0.7])
    >>> marginal_distr = MarginalDistribution(distr, d=0)
    >>> marginal_distr
    MarginalDistribution(Distribution("multinomial", n=10, p=[0.1, 0.2, 0.7]), d=0)
    >>> marginal_distr.sample(5, random_state=0).astype(int)
    array([2, 1, 2, 1, 1])
    """

    is_source_node = False

    def __init__(self, distr, d):
        self.distr = distr
        self.d = d
        super().__init__()

    def _sample(self):
        # Simply slice the parent
        return np.atleast_2d(self.distr.samples_)[:, self.d]

    def get_parents(self):
        yield self.distr

    def __repr__(self):
        return f"{type(self).__name__}({self.distr}, d={self.d})"


def MultivariateDistribution(distr, *args, **kwargs):
    """Factory function that yields marginal distributions.

    Examples
    --------
    >>> p = [0.2, 0.3, 0.5]  # Probability of each category
    >>> m1, m2, m3 = MultivariateDistribution("multinomial", n=10, p=p)
    >>> m1.sample(5, random_state=0).astype(int)
    array([3, 2, 4, 2, 1])

    Each category should sum to n=10:
    >>> (m1 + m2 + m3).sample(5, random_state=0).astype(int)
    array([10, 10, 10, 10, 10])
    """
    distr = Distribution(distr, *args, **kwargs)

    # Get dimensionality by sampling once
    d = len(distr._sample(q=[0.5]).squeeze())
    yield from (MarginalDistribution(distr, d=i) for i in range(d))


# ========================================================
if __name__ == "__main__":
    import pytest

    pytest.main(args=[__file__, "--doctest-modules", "-v", "--capture=sys"])

if __name__ == "__main__":
    rng = np.random.default_rng(42)
    a = Distribution("norm", loc=0, scale=1)
    b = Distribution("norm", loc=0, scale=1)
    c = Distribution("norm", loc=0, scale=1)

    expression = a + b
    corr_mat = np.array([[1.0, 0.8], [0.8, 1.0]])
    expression.correlate(a, b, corr_mat=corr_mat)

    expression = expression + c
    expression.correlate(b, c, corr_mat=corr_mat)

    import matplotlib.pyplot as plt

    expression.sample(999, random_state=rng)

    plt.figure(figsize=(3, 2))
    plt.scatter(a.samples_, b.samples_, s=2)
    plt.show()

    d1, d2, d3 = MultivariateDistribution("dirichlet", alpha=[1, 2, 3])

    # =========================

    cost = EmpiricalDistribution(data=[1, 2, 3, 3, 3, 3])
    norm = Distribution("norm", loc=cost, scale=1)
    (norm**2).sample(99, random_state=42)



import itertools
import numpy as np
import scipy as sp


def adjust_minmax_quantiles(quantiles, cumulatives, expected):
    """Adjust minimum and maximum in `quantiles` so we hit expected value.

    Examples
    --------
    Data for a distribution with mean (2.5 + 5.5) / 2 = 4 leads to no change:
    >>> adjust_minmax_quantiles([0, 0.5, 1], [0, 5, 6], expected=4.0)
    array([0., 5., 6.])

    Increasing the expected value to 5:
    >>> adjust_minmax_quantiles([0, 0.5, 1], [0, 5, 6], expected=5)
    array([2.00014464, 5.        , 7.99985536])

    >>> adjust_minmax_quantiles([0, 0.1, 0.3, 1], [0, 1, 1.5, 2], 1.6)
    array([0.03013972, 1.        , 1.5       , 2.20998004])
    """
    quantiles = np.array(quantiles, dtype=float)
    cumulatives = np.array(cumulatives, dtype=float)
    assert np.all(np.diff(quantiles) > 0)
    assert np.all(np.diff(cumulatives) > 0)
    assert np.isclose(np.min(quantiles), 0)
    assert np.isclose(np.max(quantiles), 1)

    def empirical_mean(quantiles, cumulatives):
        """Compute the expected value of a histogram."""
        return sp.stats.rv_histogram(
            (np.diff(quantiles), cumulatives), density=False
        ).mean()

    def transform(low_scale, high_scale, cumulatives):
        """Return new low and high values in the cumulatives."""
        cumulatives = cumulatives.copy()
        q1, q2 = cumulatives[:2]
        qn1, qn = cumulatives[-2:]
        high = max(qn1 + np.exp(high_scale) * (qn - qn1), qn1 + 1e-6)
        low = min(q2 - np.exp(low_scale) * (q2 - q1), q2 - 1e-6)
        return (low, high)

    def objective(params, quantiles, cumulatives, expected):
        """Objective function to minimize."""
        low_scale, high_scale = params

        # Transform the endpoints of the cumulatives
        (low, high) = transform(low_scale, high_scale, cumulatives)
        cumulatives_copy = np.array(cumulatives)
        cumulatives_copy[0], cumulatives_copy[-1] = low, high

        observed_mean = empirical_mean(quantiles, cumulatives_copy)

        # Create objective and return
        main_obj = np.abs(observed_mean - expected)
        bi_obj = (low - cumulatives[0]) ** 2 + (high - cumulatives[-1]) ** 2
        return main_obj + 1e-2 * bi_obj

    result = sp.optimize.minimize(
        fun=objective,
        args=(quantiles, cumulatives, expected),
        x0=[0, 0],
        method="nelder-mead",
    )
    low_scale, high_scale = result.x
    (low, high) = transform(low_scale, high_scale, cumulatives)
    cumulatives[0], cumulatives[-1] = low, high

    return cumulatives


def zip_args(args, kwargs):
    """Zip argument and keyword arguments for repeated function calls.

    Examples
    --------
    >>> args = ((1, 2, 3), itertools.repeat(None))
    >>> kwargs = {"a": (5, 6, 7), "b": itertools.repeat(9)}
    >>> for args_i, kwargs_i in zip_args(args, kwargs):
    ...     print(args_i, kwargs_i)
    (1, None) {'a': 5, 'b': 9}
    (2, None) {'a': 6, 'b': 9}
    (3, None) {'a': 7, 'b': 9}
    """
    zipped_args = zip(*args) if args else itertools.repeat(args)
    zipped_kwargs = zip(*kwargs.values()) if kwargs else itertools.repeat(kwargs)

    for args_i, kwargs_i in zip(zipped_args, zipped_kwargs):
        yield args_i, dict(zip(kwargs.keys(), kwargs_i))


def build_corrmat(correlations):
    """Given a list of [(indices1, corrmat1), (indices2, corrmat2), ...],
    create a big correlation matrix.

    Examples
    --------
    >>> correlations = [((0, 2), np.array([[1, 0.5], [0.5, 1]]))]
    >>> build_corrmat(correlations)
    array([[1. , 0. , 0.5],
           [0. , 1. , 0. ],
           [0.5, 0. , 1. ]])
    """
    # TODO: If no correlation is given, we implicitly assume zero.
    # For instance, if no correlation between indices (0, 3) is given
    # in the input data, then C[0, 3] = C[3, 0] = 0.0, which is strictly
    # speaking not the same (no preference vs. preference for 0 corr)
    n = max(max(idx) for (idx, _) in correlations)
    C = np.eye(n + 1, dtype=float)

    for idx_i, corrmat_i in correlations:
        C[np.ix_(idx_i, idx_i)] = corrmat_i

    return C


if __name__ == "__main__":
    import pytest

    pytest.main(args=[__file__, "--doctest-modules", "-v", "--capture=sys"])



import collections
from collections.abc import Collection


class GarbageCollector:
    """The GarbageCollector cleans up `.samples_` attributes as the computational
    graph is sampled. By 'garbage collection' and 'cleaning up' we mean deleting
    the `.samples_` attribute.

    Parameters
    ----------
    strategy : None or list, optional
        If None (the default), no nodes are garbage collected. If a list of
        nodes, then those nodes and the sink are NOT garbage collected. An
        empty list means all nodes except the sink will be garbage collected.
    """

    def __init__(self, strategy=None):
        if not (strategy is None or isinstance(strategy, Collection)):
            raise TypeError(f"`strategy` must be None or a collection, got: {strategy}")

        self.strategy = strategy

    def set_sink(self, sink):
        """Set the sink node, whose samples will always be kept."""
        self.sink = sink

        # The user wants to keep `.samples_` on all nodes => do nothing
        if self.strategy is None:
            return self

        # Initialize the reference counter, keeping track of the number of
        # unsampled children of all nodes. Once a node has no unsampled children
        # (i.e. all children are sampled), that node can safely be GC'ed.
        self._unsampled_children = collections.defaultdict(int)
        for node in self.sink.nodes():
            for parent in node.get_parents():
                self._unsampled_children[parent] += 1

        return self

    def decrement_and_delete(self, node):
        """Decrement the reference counter (number of unsampled children for
        each parent) and delete `.samples_` if the reference count is zero.

        Returns the nodes that were garbage collected.
        """
        if not hasattr(self, "sink"):
            raise ValueError("You must call 'set_sink' first.")

        # Nodes that were garbage collected
        garbage_collected = []

        # The user wants to keep `.samples_` on all nodes => do nothing
        if self.strategy is None:
            return []

        for parent in node.get_parents():
            # Decrement counter
            self._unsampled_children[parent] -= 1

            # Garbage collect if possible
            zero_count = self._unsampled_children[parent] == 0
            protected = parent in self.strategy
            if zero_count and not protected:
                del parent.samples_
                garbage_collected.append(parent)

            assert self._unsampled_children[parent] >= 0

        return garbage_collected



import numpy as np
import pytest
import scipy as sp
from scipy.stats import spearmanr

from probabilit.correlation import ImanConover, decorrelate


@pytest.fixture
def rng():
    return np.random.default_rng(42)


@pytest.fixture
def sample_data(rng):
    N = 1000
    K = 3
    X = np.zeros((N, K))
    X[:, 0] = rng.normal(0, 1, N)
    X[:, 1] = rng.exponential(1, N)
    X[:, 2] = rng.uniform(0, 1, N)

    # Ensure positive definite correlation matrix by constructing from random matrix
    A = rng.normal(size=(K, K))
    C = A @ A.T  # This creates a positive definite matrix
    # Convert to correlation matrix
    d = np.sqrt(np.diag(C))
    C = C / d[:, None] / d[None, :]

    return X, C


def test_preserves_marginal_distributions(rng, sample_data):
    X, C = sample_data

    X_transformed = ImanConover().set_target(C)(X)
    for k in range(X.shape[1]):
        assert np.allclose(np.sort(X[:, k]), np.sort(X_transformed[:, k]))


def test_achieves_target_correlations(sample_data):
    X, C = sample_data

    X_transformed = ImanConover().set_target(C)(X)
    rank_corr = spearmanr(X_transformed)[0]
    assert np.allclose(rank_corr, C, atol=0.05)


def test_invalid_correlation_matrix(rng):
    N, K = 100, 3
    X = rng.normal(size=(N, K))
    C_invalid = np.array(
        [
            [1.0, 0.7, -0.3],
            [0.8, 1.0, 0.5],  # Non-symmetric
            [-0.3, 0.5, 1.0],
        ]
    )

    with pytest.raises((ValueError, np.linalg.LinAlgError)):
        ImanConover().set_target(C_invalid)(X)


def test_extreme_correlations(rng):
    N, K = 1000, 3
    X = rng.normal(size=(N, K))

    # Create extreme but valid correlation matrix
    # Using the fact that a correlation matrix with ones on diagonal
    # and same value rho everywhere else is positive definite if
    # rho > -1/(K-1) where K is matrix size
    rho = 0.99  # High correlation but still allows matrix to be positive definite
    C_extreme = np.ones((K, K)) * rho
    np.fill_diagonal(C_extreme, 1.0)

    # Verify matrix is positive definite before test
    eigenvals = np.linalg.eigvals(C_extreme)
    assert np.all(eigenvals > 0), "Test correlation matrix is not positive definite"

    X_extreme = ImanConover().set_target(C_extreme)(X)
    rank_corr_extreme = spearmanr(X_extreme)[0]
    assert np.allclose(rank_corr_extreme, C_extreme, atol=0.05)


def test_correlation_matrix_validation(rng):
    N = 100
    K = 3
    X = rng.normal(size=(N, K))

    # Create non-positive definite matrix
    C_invalid = np.array([[1.0, 2.0, 0.3], [2.0, 1.0, 0.2], [0.3, 0.2, 1.0]])

    with pytest.raises((ValueError, np.linalg.LinAlgError)):
        ImanConover().set_target(C_invalid)(X)


def test_input_validation(rng):
    N = 100
    K = 3
    X = rng.normal(size=(N, K))
    C = np.array(
        [
            [1.0, 0.5],  # Wrong size
            [0.5, 1.0],
        ]
    )

    with pytest.raises(ValueError):
        ImanConover().set_target(C)(X)


def test_orthogonality_precision(rng):
    """Test that uncorrelated variables remain very close to orthogonal"""
    N = 100
    K = 4

    # Create target correlation matrix with some zeros
    C_target = np.array(
        [
            [1.0, 0.0, 0.5, 0.0],
            [0.0, 1.0, 0.0, 0.5],
            [0.5, 0.0, 1.0, 0.0],
            [0.0, 0.5, 0.0, 1.0],
        ]
    )

    X = rng.normal(size=(N, K))
    X_transformed = ImanConover().set_target(C_target)(X)
    rank_corr = spearmanr(X_transformed)[0]

    # Get the elements that should be zero
    zero_mask = C_target == 0
    achieved_zeros = rank_corr[zero_mask]

    # With variance reduction, these should be very close to zero
    # Using much tighter tolerance than regular correlation tests
    # Threshold found by running thousands of tests.
    # This will fail much more often if we do not use the variance
    # reduction technique described in the paper.
    assert np.all(np.abs(achieved_zeros) < 0.12), (
        "Zero correlations not maintained with sufficient precision"
    )


class TestImanConover:
    @pytest.mark.parametrize("seed", range(100))
    def test_marginals_and_correlation_distance(self, seed):
        rng = np.random.default_rng(seed)

        n_variables = rng.integers(2, 100)
        n_observations = n_variables * 10

        # Create a random correlation matrix and a random data matrix
        A = rng.normal(size=(n_variables * 2, n_variables))
        desired_corr = 0.9 * np.corrcoef(A, rowvar=False) + 0.1 * np.eye(n_variables)
        X = rng.normal(size=(n_observations, n_variables))

        # Tranform the data
        transform = ImanConover().set_target(desired_corr)
        X_transformed = transform(X)

        # Check that all columns (variables) have equal marginals.
        # In other words, Iman-Conover can permute each column individually,
        # but they should have identical entries before and after.
        for j in range(X.shape[1]):
            assert np.allclose(np.sort(X[:, j]), np.sort(X_transformed[:, j]))

        # After the Iman-Conover transform, the distance between the desired
        # correlation matrix should be smaller than it was before.
        X_corr = np.corrcoef(X, rowvar=False)
        distance_before = sp.linalg.norm(X_corr - desired_corr, ord="fro")

        X_trans_corr = np.corrcoef(X_transformed, rowvar=False)
        distance_after = sp.linalg.norm(X_trans_corr - desired_corr, ord="fro")

        assert distance_after <= distance_before

    def test_identity_correlation_matrix(self):
        rng = np.random.default_rng(42)

        n_observations = 5
        n_variables = 3
        rng = np.random.default_rng(42)

        # Create a random correlation matrix and a random data matrix
        desired_corr = np.identity(n_variables)
        transform = ImanConover().set_target(desired_corr)

        # Create data and decorrelate it completely
        X = rng.normal(size=(n_observations, n_variables))
        X = decorrelate(X, remove_variance=True)
        assert np.allclose(np.corrcoef(X, rowvar=False), np.eye(n_variables))

        # Transform it to identity correlation, which it already has
        transform = ImanConover().set_target(desired_corr)
        X_transformed = transform(X)

        assert np.allclose(X, X_transformed)

    def test_dataset_with_unity_correlation_in_ranks(self):
        # This dataset is interesting because while the correlation
        # between the variables is ~0.6, when the data is ranked the
        # correlation becomes 1. Rank(row) = [1, 2, 3] for both rows.
        X = np.array([[1.0, 1], [2.0, 1.1], [2.1, 3]])

        desired_corr = np.identity(2)

        transform = ImanConover().set_target(desired_corr)
        with pytest.raises(ValueError):
            transform(X)


if __name__ == "__main__":
    pytest.main(args=[__file__, "--doctest-modules", "-v", "-l"])



from probabilit.distributions import (
    _fit_triangular_distribution,
    _pert_to_beta,
    Lognormal,
    Uniform,
)
import pytest
from scipy.stats import triang
import numpy as np


class TestTriangular:
    @pytest.mark.parametrize("c", [0.1, 0.5, 0.7])
    @pytest.mark.parametrize("loc", [-1, 0, 1])
    @pytest.mark.parametrize("scale", [1, 10, 25])
    @pytest.mark.parametrize("low_perc", [0.01, 0.05, 0.1, 0.2])
    def test_triangular_roundstrips(self, c, loc, scale, low_perc):
        # Test round-trips
        mode = loc + c * scale
        high_perc = 0.8

        # Get parameters to optimize toward
        distr = triang(loc=loc, scale=scale, c=c)
        target_low, target_high = distr.ppf([low_perc, high_perc])

        # Found parameters
        loc_f, scale_f, c_f = _fit_triangular_distribution(
            mode=mode,
            low=target_low,
            high=target_high,
            low_perc=low_perc,
            high_perc=high_perc,
        )

        np.testing.assert_allclose([loc_f, scale_f, c_f], [loc, scale, c], atol=1e-8)

    @pytest.mark.parametrize("delta", [0.001, 0.01, 0.1, 0.2, 0.3, 0.4])
    def test_triangular_roundstrips_squeeze(self, delta):
        loc = 0
        scale = 10
        c = 0.8
        mode = loc + c * scale

        # Get parameters to optimize toward
        distr = triang(loc=loc, scale=scale, c=c)
        target_low, target_high = distr.ppf([delta, 1 - delta])

        # Found parameters
        loc_f, scale_f, c_f = _fit_triangular_distribution(
            mode=mode,
            low=target_low,
            high=target_high,
            low_perc=delta,
            high_perc=1 - delta,
        )

        np.testing.assert_allclose([loc_f, scale_f, c_f], [loc, scale, c], atol=1e-8)


class TestLognormal:
    @pytest.mark.parametrize(
        "mean,std",
        [
            (1.0, 0.5),
            (10.0, 1.0),
            (100.0, 20.0),
        ],
    )
    def test_lognormal_moments(self, mean, std):
        rng = np.random.default_rng(42)
        dist = Lognormal(mean, std)
        samples = dist.sample(10000, random_state=rng)

        np.testing.assert_allclose(np.mean(samples), mean, rtol=0.05)
        np.testing.assert_allclose(np.std(samples), std, rtol=0.05)

    @pytest.mark.parametrize(
        "mu,sigma",
        [
            (0.0, 0.5),
            (1.0, 1.0),
            (-0.5, 0.3),
        ],
    )
    def test_lognormal_from_log_params_moments(self, mu, sigma):
        rng = np.random.default_rng(42)
        dist = Lognormal.from_log_params(mu, sigma)
        samples = dist.sample(10000, random_state=rng)

        # Calculate expected moments from log-space parameters
        expected_mean = np.exp(mu + sigma**2 / 2)
        expected_variance = (np.exp(sigma**2) - 1) * np.exp(2 * mu + sigma**2)
        expected_std = np.sqrt(expected_variance)

        np.testing.assert_allclose(np.mean(samples), expected_mean, rtol=0.05)
        np.testing.assert_allclose(np.std(samples), expected_std, rtol=0.05)


class TestPERT:
    @pytest.mark.parametrize("gamma", [1, 3, 4, 7])
    @pytest.mark.parametrize("maximum", [10, 12, 14])
    def test_pert_properties(self, gamma, maximum):
        # Convert from PERT parameters to beta
        a, b, loc, scale = _pert_to_beta(
            minimum=1, mode=4, maximum=maximum, gamma=gamma
        )

        # The mode of the beta distribution (from Wikipedia)
        mode = (a - 1) / (a + b - 2)
        # The mode should be located in the correct positoin on [0, 1]
        np.testing.assert_allclose(mode, (4 - 1) / (maximum - 1))

        # Desired mean of PERT matches actual mean of beta
        mean = (1 + gamma * 4 + maximum) / (gamma + 2)
        np.testing.assert_allclose(mean, (a / (a + b)) * scale + loc)


class TestUniform:
    @pytest.mark.parametrize(
        "min_val,max_val", [(-5, 5), (0, 15), (10, 100), (-10, -2), (1, 2)]
    )
    def test_uniform_properties(self, min_val, max_val):
        rng = np.random.default_rng(42)
        dist = Uniform(min_val, max_val)
        samples = dist.sample(10000, random_state=rng)

        # Test that all samples are within bounds
        assert np.all(samples >= min_val)
        assert np.all(samples <= max_val)

        expected_mean = (min_val + max_val) / 2
        np.testing.assert_allclose(np.mean(samples), expected_mean, rtol=0.1, atol=0.1)

        expected_variance = (max_val - min_val) ** 2 / 12
        expected_std = np.sqrt(expected_variance)
        np.testing.assert_allclose(np.std(samples), expected_std, rtol=0.05)


if __name__ == "__main__":
    import pytest

    pytest.main(args=[__file__, "--doctest-modules", "-v", "--capture=sys"])



"""
Inspection
----------

Inspection of results, plotting, tables, exporting, etc.
"""

import seaborn
import pandas as pd
from probabilit.modeling import NoOp, Distribution, Transform
import numpy as np
from numbers import Number


def plot(*variables, corr=None, sample_kwargs=None, **kwargs):
    """Utility function for quick plotting of one or several variables.

    Examples
    --------
    >>> a = Distribution("uniform", loc=0, scale=1)
    >>> b = Distribution("uniform", loc=0, scale=1)
    >>> c = Distribution("uniform", loc=0, scale=1)

    >>> pairgrid = plot(a)
    >>> pairgrid = plot(a, b)
    >>> pairgrid = plot(a, b, corr=0.5)

    >>> corr = np.eye(3) / 2 + np.ones((3, 3)) / 2
    >>> pairgrid = plot(a, b, c, corr=corr)

    >>> pairgrid = plot(a, sample_kwargs={'size':99})
    """
    # Create an NoOp node, then copy the NoOp (which copies all parents too)
    # This prevents us from mutating the input arguments
    no_operation = NoOp(*variables).copy()
    variables = no_operation.parents

    if len(variables) == 2 and isinstance(corr, Number):
        corr = np.array([[1.0, corr], [corr, 1.0]])

    # Check if variables are already sampled
    sampled = [hasattr(v, "samples_") for v in variables]

    if any(sampled) and not all(sampled):
        raise ValueError("Either all variables must be sampled, or none.")

    # Sample if not sampled, or any  keyword args are specified
    if not any(sampled) or (corr is not None) or (sample_kwargs is not None):
        # Apply defaults first
        sample_kwargs = {"size": 999, "random_state": 0} | (sample_kwargs or {})

        # Correlate if a correlation is given
        if corr is not None:
            no_operation.correlate(*variables, corr_mat=corr)

        no_operation.sample(**sample_kwargs)

    # Transform to dataframe and return plot
    df = pd.DataFrame(
        {f"var_{i}": var.samples_ for (i, var) in enumerate(variables, 1)}
    )
    return seaborn.pairplot(df, **kwargs)


def treeprint(node):
    """Print a computational graph in a tree-like fashion.

    Examples
    --------
    >>> scale = Distribution("expon")
    >>> a = Distribution("norm", loc=1, scale=scale)
    >>> treeprint(a + scale - scale**2)
    Subtract
       ├──Add
       │  ├──Distribution("norm", loc=1, scale=Distribution("expon"))
       │  │  └──Distribution("expon")
       │  └──Distribution("expon")
       └──Power
          ├──Distribution("expon")
          └──Constant(2)

    """
    elbow, pipe, tee, blank = "└──", "│  ", "├──", "   "

    def _treeprint(node, last=True, header="", root=False):
        # Recursive version
        output = type(node).__name__ if isinstance(node, Transform) else str(node)
        print(header + ("" if root else (elbow if last else tee)) + output)

        if parents := list(node.get_parents()):
            for i, parent in enumerate(parents):
                _treeprint(
                    parent,
                    header=header + (blank if last else pipe),
                    last=i == len(parents) - 1,
                )

    return _treeprint(node, last=True, header="", root=True)


if __name__ == "__main__":
    rng = np.random.default_rng(42)
    # mu = Constant(1)
    a = Distribution("norm", loc=0, scale=1)
    b = Distribution("norm", loc=a, scale=0.5)

    plot(a, b)
    grid = plot(a)

    from probabilit.modeling import MultivariateDistribution

    cov = np.array([[1, 0.9], [0.9, 1]])
    n1, n2 = MultivariateDistribution("multivariate_normal", mean=[1, 2], cov=cov)
    from probabilit.inspection import plot

    plot(n1, n2)



from probabilit.modeling import (
    Distribution,
    Constant,
    EmpiricalDistribution,
    CumulativeDistribution,
    DiscreteDistribution,
    Equal,
    Exp,
    Log,
    Min,
    Max,
    All,
    Any,
    scalar_transform,
    MultivariateDistribution,
)
from probabilit.distributions import (
    PERT,
    Triangular,
    Normal,
    Lognormal,
    Uniform,
    TruncatedNormal,
)
from probabilit.inspection import plot, treeprint


__all__ = [
    # Geneal modeling
    "Distribution",
    "Constant",
    "EmpiricalDistribution",
    "CumulativeDistribution",
    "DiscreteDistribution",
    "Equal",
    "Exp",
    "Log",
    "Min",
    "Max",
    "All",
    "Any",
    "scalar_transform",
    "MultivariateDistribution",
    # Custom distributions
    "PERT",
    "Triangular",
    "Normal",
    "Lognormal",
    "Uniform",
    "TruncatedNormal",
    # Inspection
    "plot",
    "treeprint",
]



from probabilit.correlation import (
    Permutation,
    CorrelationMatrix,
    SwapIndexGenerator,
)
import numpy as np
import pytest
import scipy as sp


class TestSwapIndexGenerator:
    @pytest.mark.parametrize("seed", range(10))
    def test_disjoint(self, seed):
        rng = np.random.default_rng(seed)
        n = rng.integers(5, 15)
        generator = SwapIndexGenerator(rng=rng, n=n)

        for _ in range(50):
            size = rng.integers(1, 20)
            i, j = generator(size)
            assert not set(i).intersection(set(j))


class TestCorrelationMatrix:
    @pytest.mark.parametrize("seed", range(10))
    @pytest.mark.parametrize("correlation_type", ["pearson", "spearman"])
    def test_squences_of_swaps(self, seed, correlation_type):
        def corr(X, correlation_type):
            if correlation_type == "pearson":
                return np.corrcoef(X, rowvar=False)
            elif correlation_type == "spearman":
                return sp.stats.spearmanr(X).statistic
            else:
                raise ValueError("Invalid correlation type")

        rng = np.random.default_rng(seed)
        X = rng.normal(size=(9, 4))

        correlation_matrix = CorrelationMatrix(X, correlation_type=correlation_type)

        # Test that the CorrelationMatrix computes correlation correctly
        observed_corr = corr(X, correlation_type)
        np.testing.assert_allclose(observed_corr, correlation_matrix[:, :], atol=1e-12)

        # Test two consequtive single swap
        # --------------------------------
        X_swapped = np.copy(X)
        correlation_matrix.commit(col=1, i=0, j=1)
        X_swapped[0, 1], X_swapped[1, 1] = X_swapped[1, 1], X_swapped[0, 1]

        observed_corr = corr(X_swapped, correlation_type)
        np.testing.assert_allclose(observed_corr, correlation_matrix[:, :], atol=1e-12)

        # Test another swap on the same matrix
        correlation_matrix.commit(col=1, i=2, j=3)
        X_swapped[2, 1], X_swapped[3, 1] = X_swapped[3, 1], X_swapped[2, 1]

        observed_corr = corr(X_swapped, correlation_type)
        np.testing.assert_allclose(observed_corr, correlation_matrix[:, :], atol=1e-12)

        # Test two swaps on the same matrix at once
        # -----------------------------------------
        correlation_matrix = CorrelationMatrix(X, correlation_type=correlation_type)
        X_swp = np.copy(X)
        correlation_matrix.commit(col=1, i=[0, 1], j=[2, 3])
        X_swp[[0, 1], 1], X_swp[[2, 3], 1] = X_swp[[2, 3], 1], X_swp[[0, 1], 1]

        observed_corr = corr(X_swp, correlation_type)
        np.testing.assert_allclose(observed_corr, correlation_matrix[:, :], atol=1e-12)

        # Check that committing back and forth leads to the same matrix
        before = np.copy(correlation_matrix.X)
        correlation_matrix.commit(col=1, i=[0, 1], j=[2, 3])
        midway = np.copy(correlation_matrix.X)
        correlation_matrix.commit(col=1, i=[0, 1], j=[2, 3])
        after = np.copy(correlation_matrix.X)

        np.testing.assert_allclose(before, after)
        assert not np.allclose(before, midway)

        # Test a single swap
        a, b = correlation_matrix.X[0, 2], correlation_matrix.X[5, 2]
        correlation_matrix.commit(col=2, i=0, j=5)
        np.testing.assert_allclose(a, correlation_matrix.X[5, 2])
        np.testing.assert_allclose(b, correlation_matrix.X[0, 2])

    @pytest.mark.parametrize("seed", range(100))
    def test_swap(self, seed):
        rng = np.random.default_rng(seed)
        X = rng.normal(size=(9, 4))

        # Test a single swap
        correlation_matrix = CorrelationMatrix(X)
        a, b = correlation_matrix.X[0, 2], correlation_matrix.X[5, 2]
        correlation_matrix.commit(col=2, i=0, j=5)
        np.testing.assert_allclose(a, correlation_matrix.X[5, 2])
        np.testing.assert_allclose(b, correlation_matrix.X[0, 2])

    @pytest.mark.parametrize("seed", range(100))
    @pytest.mark.parametrize("correlation_type", ["pearson", "spearman"])
    def test_that_many_single_equals_one_large_swap(self, seed, correlation_type):
        rng = np.random.default_rng(seed)
        X = rng.normal(size=(9, 4))

        # Test a chain of swaps
        correlation_matrix = CorrelationMatrix(X, correlation_type=correlation_type)
        a, b = [1, 3, 2], [6, 4, 5]
        for i, j in zip(a, b):
            correlation_matrix.commit(col=2, i=i, j=j)
        X_singles = np.copy(correlation_matrix.X)

        correlation_matrix = CorrelationMatrix(X, correlation_type=correlation_type)
        correlation_matrix.commit(col=2, i=a, j=b)
        X_multiples = np.copy(correlation_matrix.X)

        np.testing.assert_allclose(X_singles, X_multiples)


class TestPermutation:
    @pytest.mark.parametrize("seed", range(25))
    def test_convergence(self, seed):
        # With default parameters, convergence should be decent
        # regardless of the size of the problem
        rng = np.random.default_rng(seed)

        n_variables = rng.integers(2, 20)
        n_observations = n_variables * rng.integers(5, 50)

        # Create a correlation matrix and a random data matrix
        desired_corr = np.ones((n_variables, n_variables)) * 0.5
        np.fill_diagonal(desired_corr, val=1.0)
        X = rng.normal(size=(n_observations, n_variables))

        # Tranform the data
        transform = Permutation(random_state=seed).set_target(desired_corr)
        X_transformed = transform(X)

        actual_corr = np.corrcoef(X_transformed, rowvar=False)
        rmse = np.sqrt(np.mean((actual_corr - desired_corr) ** 2))
        assert rmse < 0.1

    @pytest.mark.parametrize("seed", range(10))
    def test_marginals_and_correlation_distance(self, seed):
        rng = np.random.default_rng(seed)

        n_variables = rng.integers(2, 10)
        n_observations = n_variables * 10

        # Create a random correlation matrix and a random data matrix
        A = rng.normal(size=(n_variables * 2, n_variables))
        desired_corr = 0.9 * np.corrcoef(A, rowvar=False) + 0.1 * np.eye(n_variables)
        X = rng.normal(size=(n_observations, n_variables))

        # Tranform the data
        transform = Permutation(random_state=0, iterations=10)
        transform = transform.set_target(desired_corr)
        X_transformed = transform(X)

        # Check that all columns (variables) have equal marginals.
        for j in range(X.shape[1]):
            assert np.allclose(np.sort(X[:, j]), np.sort(X_transformed[:, j]))

        # After using the Permutation, the distance to the
        # desired correlation matrix should be smaller than it was before.
        X_corr = np.corrcoef(X, rowvar=False)
        distance_before = sp.linalg.norm(X_corr - desired_corr, ord="fro")

        X_trans_corr = np.corrcoef(X_transformed, rowvar=False)
        distance_after = sp.linalg.norm(X_trans_corr - desired_corr, ord="fro")

        assert distance_after <= distance_before

    def test_dataset_with_more_variables_than_observations(self):
        rng = np.random.default_rng(42)
        X = rng.normal(size=(5, 10))

        desired_corr = np.identity(10)
        transform = Permutation(random_state=0).set_target(desired_corr)
        X_trans = transform(X)

        cor_X_trans = np.corrcoef(X_trans, rowvar=False)
        cor_X = np.corrcoef(X, rowvar=False)
        assert transform._error(cor_X_trans, desired_corr) < transform._error(
            cor_X, desired_corr
        )


if __name__ == "__main__":
    pytest.main(args=[__file__, "--doctest-modules", "-v", "-l"])



"""
Correlation
-----------

Inducing correlations, working with correlation matrices, etc.


Using Iman-Conover with Latin Hybercube sampling
------------------------------------------------

Sample on the unit hypercube using LatinHypercube

>>> import scipy as sp
>>> sampler = sp.stats.qmc.LatinHypercube(d=2, seed=42, scramble=True)
>>> samples = sampler.random(n=100)

Map to distributions

>>> X = np.vstack((sp.stats.triang(0.5).ppf(samples[:, 0]),
...                sp.stats.gamma.ppf(samples[:, 1], a=1))).T

Induce correlations

>>> float(sp.stats.pearsonr(*X.T).statistic)
0.065898...
>>> correlation_matrix = np.array([[1, 0.3], [0.3, 1]])
>>> transform = ImanConover().set_target(correlation_matrix)
>>> X_transformed = transform(X)
>>> float(sp.stats.pearsonr(*X_transformed.T).statistic)
0.279652...
"""

import numpy as np
import scipy as sp
import abc
import dataclasses

import contextlib
import os
import itertools

# CVXPY prints error messages about incompatible ortools version during import.
# Since we use the SCS solver and not GLOP/PDLP (which need ortools), these errors
# are irrelevant and would only confuse users. We suppress them by redirecting
# stdout/stderr during import.
# https://github.com/cvxpy/cvxpy/issues/2470
with (
    open(os.devnull, "w") as devnull,
    contextlib.redirect_stdout(devnull),
    contextlib.redirect_stderr(devnull),
):
    import cvxpy as cp


class CorrelatorError(Exception):
    pass


def nearest_correlation_matrix(matrix, *, weights=None, eps=1e-6, verbose=False):
    """Returns the correlation matrix nearest to `matrix`, weighted elementwise
    by `weights`.

    Parameters
    ----------
    matrix : np.ndarray
        The matrix that we want to find the nearest correlation matrix to.
        A square 2-dimensional NumPy ndarray.
    weights : np.ndarray or None, optional
        An elementwise weighting matrix. A square 2-dimensional NumPy ndarray
        that must have the same shape as `matrix`. The default is None.
    eps : float, optional
        Tolerance for the optimization solver. The default is 1e-6.
    verbose : bool, optional
        Whether to print information from the solver. The default is False.

    Returns
    -------
    np.ndarray
        The correlation matrix that is nearest to the input matrix.

    Notes
    -----
    This function implements equation (3) in the paper "An Augmented Lagrangian
    Dual Approach for the H-Weighted Nearest Correlation Matrix Problem" by
    Houduo Qi and Defeng Sun.
        http://www.personal.soton.ac.uk/hdqi/REPORTS/Cor_matrix_H.pdf
    Another useful link is:
        https://nhigham.com/2020/04/14/what-is-a-correlation-matrix/

    Examples
    --------
    >>> X = np.array([[1, 1, 0],
    ...               [1, 1, 1],
    ...               [0, 1, 1]])
    >>> nearest_correlation_matrix(X)
    array([[1.        , 0.76068..., 0.15729...],
           [0.76068..., 1.        , 0.76068...],
           [0.15729..., 0.76068..., 1.        ]])
    >>> H = np.array([[1,   0.5, 0.1],
    ...               [0.5,   1, 0.5],
    ...               [0.1, 0.5, 1]])
    >>> nearest_correlation_matrix(X, weights=H)
    array([[1.        , 0.94171..., 0.77365...],
           [0.94171..., 1.        , 0.94171...],
           [0.77365..., 0.94171..., 1.        ]])
    """
    if not isinstance(matrix, np.ndarray):
        raise TypeError("Input argument `matrix` must be np.ndarray.")
    if not matrix.ndim == 2 and matrix.shape[0] == matrix.shape[1]:
        raise ValueError("Input argument `matrix` must be square.")

    # If the matrix is already valid, return it back immediately
    valid_diag = np.allclose(np.diag(matrix), 1)
    valid_entries = np.allclose(np.clip(matrix, -1, 1), matrix)
    if all((valid_diag, valid_entries, _is_positive_definite(matrix))):
        return matrix

    # Switch to notation used in the paper
    G = matrix.copy()
    H = np.ones_like(G) if weights is None else weights

    if not isinstance(H, np.ndarray):
        raise TypeError("Input argument `weights` must be np.ndarray.")
    if not (H.shape == G.shape):
        raise ValueError("Argument `weights` must have same shape as `matrix`.")

    # To constrain Y to be Positive Symmetric Definite (PSD), you need to
    # either set PSD=True here, or add the special constraint 'Y >> 0'. See:
    # https://www.cvxpy.org/tutorial/constraints/index.html#semidefinite-matrices
    X = cp.Variable(shape=G.shape, PSD=True)

    # Objective and constraints for minimizing the weighted frobenius norm.
    # This is equation (3) in the paper. We set (X - eps * I) >> 0 as an extra
    # constraint. Mathematically this is not needed, but numerically it helps
    # by nudging the solution slightly more, so the minimum eigenvalue is > 0.
    objective = cp.norm(cp.multiply(H, X - G), "fro")
    eps_identity = (eps / G.shape[0]) * 10
    constraints = [cp.diag(X) == 1.0, (X - eps_identity * np.eye(G.shape[0])) >> 0]

    # For solver options, see:
    # https://www.cvxpy.org/tutorial/solvers/index.html#setting-solver-options
    problem = cp.Problem(cp.Minimize(objective), constraints)
    problem.solve(solver="SCS", verbose=verbose, eps=eps)
    X = X.value.copy()  # Copy over solution

    # We might get small eigenvalues due to numerics. Attempt to fix this by
    # recursively calling the solver with smaller values of epsilon. This is
    # an extra fail-safe that is very rarely triggered on actual data.
    is_symmetric = np.allclose(X, X.T)
    is_PD = np.linalg.eig(X)[0].min() > 0
    if not (is_symmetric and is_PD) and (eps > 1e-14):
        if verbose:
            print(f"Recursively calling solver with eps := {eps} / 10")
        return nearest_correlation_matrix(G, weights=H, eps=eps / 10, verbose=verbose)

    return X


def _is_positive_definite(X):
    try:
        np.linalg.cholesky(X)
        return True
    except np.linalg.LinAlgError:
        return False


class Correlator(abc.ABC):
    def set_target(self, correlation_matrix, cholesky=True):
        """Set target correlation matrix."""
        if not isinstance(correlation_matrix, np.ndarray):
            raise TypeError("Input argument `correlation_matrix` must be NumPy array.")
        if not correlation_matrix.ndim == 2:
            raise ValueError("Correlation matrix must be square.")
        if not correlation_matrix.shape[0] == correlation_matrix.shape[1]:
            raise ValueError("Correlation matrix must be square.")
        if not np.allclose(np.diag(correlation_matrix), 1.0):
            raise ValueError("Correlation matrix must have 1.0 on diagonal.")
        if not np.allclose(correlation_matrix.T, correlation_matrix):
            raise ValueError("Correlation matrix must be symmetric.")
        if not _is_positive_definite(correlation_matrix):
            raise ValueError("Correlation matrix must be positive definite.")

        self.C = correlation_matrix.copy()
        if cholesky:
            self.P = np.linalg.cholesky(self.C)
        return self

    def _validate_X(self, X, check_rows_cols=True):
        """Validate array X of shape (observations, variables)."""
        if not (hasattr(self, "C")):
            raise CorrelatorError("User must call `set_target` first.")

        if not isinstance(X, np.ndarray):
            raise TypeError("Input argument `X` must be NumPy array.")
        if not X.ndim == 2:
            raise ValueError("Correlation matrix must be square.")

        N, K = X.shape

        if self.C.shape[0] != K:
            msg = f"Shape of `X` ({X.shape}) does not match shape of "
            msg += f"correlation matrix ({self.P.shape})"
            raise ValueError(msg)

        if check_rows_cols and N <= K:
            msg = f"The matrix X must have rows > columns. Got shape: {X.shape}"
            raise ValueError(msg)

        return N, K


@dataclasses.dataclass(init=False, repr=True, eq=False)
class Cholesky(Correlator):
    """Create a Cholesky transform.

    Parameters
    ----------
    correlation_matrix : ndarray
        Target correlation matrix of shape (K, K). The Iman-Conover will
        try to induce a correlation on the data set X so that corr(X) is
        as close to `correlation_matrix` as possible.

    Examples
    --------
    Create a desired correction of 0.7 and a data set X with no correlation.
    >>> correlation_matrix = np.array([[1, 0.7], [0.7, 1]])
    >>> rng = np.random.default_rng(4)
    >>> X = rng.normal(size=(9, 2))
    >>> sp.stats.pearsonr(*X.T).statistic.round(6)
    np.float64(-0.025582)

    >>> transform = Cholesky().set_target(correlation_matrix)
    >>> X_transformed = transform(X)
    >>> sp.stats.pearsonr(*X_transformed.T).statistic.round(6)
    np.float64(0.7)

    Verify that mean and std is the same before and after:
    >>> np.mean(X, axis=0)
    array([-0.63531692,  0.70114825])
    >>> np.mean(X_transformed, axis=0)
    array([-0.63531692,  0.70114825])
    >>> np.std(X, axis=0)
    array([1.11972638, 0.75668173])
    >>> np.std(X_transformed, axis=0)
    array([1.11972638, 0.75668173])

    """

    def set_target(self, correlation_matrix):
        super().set_target(correlation_matrix)
        return self

    def __call__(self, X):
        """Transform an input matrix X.

        Parameters
        ----------
        X : ndarray
            Input matrix of shape (N, K). This is the data set that we want to
            induce correlation structure on. X must have at least K + 1
            independent rows, because corr(X) cannot be singular.

        Returns
        -------
        ndarray
            Output matrix of shape (N, K). This data set will have a
            correlation structure that is more similar to `correlation_matrix`.

        """
        self._validate_X(X)
        N, K = X.shape

        # Remove existing mean and std from marginal distributions
        mean = np.mean(X, axis=0)
        std = np.std(X, axis=0)
        X_n = (X - mean) / std

        # Removing existing correlation
        cov = np.cov(X_n, rowvar=False, ddof=0)
        P = np.linalg.cholesky(cov)  # P is lower triangular

        # Compute X_n = X_n @ inv(P).T @ self.P.T
        # Several numerical improvements are available to us here:
        # (1) Evaluate left-to-right or right-to-left depending on sizes
        # (2) Do not invert the matrix, instead solve triangular systems
        # (3) Use LAPACK routine TRMM (scipy.linalg.blas.dtrmm)
        # I choose to implement (1) and (2), but avoid the LACACK calls for now.

        # When it comes to evaluation order (point (1) above), it's better to
        # evaluate left-to-right if N < K, and right-to-left if N > K.
        # Since N > K (must have rows > columns), we evaluate right-to-left
        transform = sp.linalg.solve_triangular(P.T, self.P.T, lower=False)
        return mean + X_n @ (transform * std)


@dataclasses.dataclass(init=False, repr=True, eq=False)
class ImanConover(Correlator):
    """Create an Iman-Conover transform.

    Parameters
    ----------
    correlation_matrix : ndarray
        Target correlation matrix of shape (K, K). The Iman-Conover will
        try to induce a correlation on the data set X so that corr(X) is
        as close to `correlation_matrix` as possible.

    Notes
    -----
    The implementation follows the original paper:
    Iman, R. L., & Conover, W. J. (1982). A distribution-free approach to
    inducing rank correlation among input variables. Communications in
    Statistics - Simulation and Computation, 11(3), 311-334.
    https://www.tandfonline.com/doi/epdf/10.1080/03610918208812265?needAccess=true
    https://www.uio.no/studier/emner/matnat/math/STK4400/v05/undervisningsmateriale/A%20distribution-free%20approach%20to%20rank%20correlation.pdf

    Other useful sources:
    - https://blogs.sas.com/content/iml/2021/06/16/geometry-iman-conover-transformation.html
    - https://blogs.sas.com/content/iml/2021/06/14/simulate-iman-conover-transformation.html
    - https://aggregate.readthedocs.io/en/stable/5_technical_guides/5_x_iman_conover.html

    Examples
    --------
    Create a desired correction of 0.7 and a data set X with no correlation.
    >>> correlation_matrix = np.array([[1, 0.7], [0.7, 1]])
    >>> transform = ImanConover().set_target(correlation_matrix)
    >>> X = np.array([[0, 0  ],
    ...               [0, 0.5],
    ...               [0,  1 ],
    ...               [1, 0  ],
    ...               [1, 0.5],
    ...               [1, 1  ]])
    >>> X_transformed = transform(X)
    >>> X_transformed
    array([[0. , 0. ],
           [0. , 0. ],
           [0. , 0.5],
           [1. , 0.5],
           [1. , 1. ],
           [1. , 1. ]])

    The original data X has no correlation at all, while the transformed
    data has correlation that is closer to the desired correlation structure:

    >>> sp.stats.pearsonr(*X.T).statistic.round(6)
    np.float64(0.0)
    >>> sp.stats.pearsonr(*X_transformed.T).statistic.round(6)
    np.float64(0.816497)

    Achieving the exact correlation structure might be impossible. For the
    input matrix above, there is no permutation of the columns that yields
    the exact desired correlation of 0.7. Iman-Conover is a heuristic that
    tries to get as close as possible.

    With many samples, we get good results if the data are normal:

    >>> rng = np.random.default_rng(42)
    >>> X = rng.normal(size=(1000, 2))
    >>> X_transformed = transform(X)
    >>> sp.stats.pearsonr(*X_transformed.T).statistic.round(6)
    np.float64(0.697701)

    But if the data are far from normal (here:lognormal), the results are
    not as good. This is because correlation is induced in a normal space
    before the result is mapped back to the original marginal distributions.

    >>> rng = np.random.default_rng(42)
    >>> X = rng.lognormal(size=(1000, 2))
    >>> X_transformed = transform(X)
    >>> sp.stats.pearsonr(*X_transformed.T).statistic.round(6)
    np.float64(0.592541)
    """

    def set_target(self, correlation_matrix):
        super().set_target(correlation_matrix)
        return self

    def __call__(self, X):
        """Transform an input matrix X.

        The output will have the same marginal distributions, but with
        induced correlation.

        Parameters
        ----------
        X : ndarray
            Input matrix of shape (N, K). This is the data set that we want to
            induce correlation structure on. X must have at least K + 1
            independent rows, because corr(X) cannot be singular.

        Returns
        -------
        ndarray
            Output matrix of shape (N, K). This data set will have a
            correlation structure that is more similar to `correlation_matrix`.

        """
        self._validate_X(X)
        N, K = X.shape

        # STEP ONE - Use van der Waerden scores to transform data to
        # approximately multivariate normal (but with correlations).
        # The new data has the same rank correlation as the original data.
        ranks = sp.stats.rankdata(X, axis=0) / (N + 1)
        normal_scores = sp.stats.norm.ppf(ranks)  # + np.random.randn(N, K) * epsilon

        # STEP TWO - Remove correlations from the transformed data
        empirical_correlation = np.corrcoef(normal_scores, rowvar=False)
        if not _is_positive_definite(empirical_correlation):
            msg = "Rank data correlation not positive definite."
            msg += "There are perfect correlations in the ranked data."
            msg += "Supply more data (rows in X) or sample differently."
            raise ValueError(msg)

        decorrelation_matrix = np.linalg.cholesky(empirical_correlation)

        # We exploit the fact that Q is lower-triangular and avoid the inverse.
        # X = N @ inv(Q)^T  =>  X @ Q^T = N  =>  (Q @ X^T)^T = N
        decorrelated_scores = sp.linalg.solve_triangular(
            decorrelation_matrix, normal_scores.T, lower=True
        ).T

        # STEP THREE - Induce correlations in transformed space
        correlated_scores = decorrelated_scores @ self.P.T

        # STEP FOUR - Map back to original space using ranks, ensuring
        # that marginal distributions are preserved
        result = np.empty_like(X)
        for k in range(K):
            # If row j is the k'th largest in `correlated_scores`, then
            # we map the k'th largest entry in X to row j.
            ranks = sp.stats.rankdata(correlated_scores[:, k]).astype(int) - 1
            result[:, k] = np.sort(X[:, k])[ranks]

        return result


@dataclasses.dataclass
class SwapIndexGenerator:
    """Returns tuples of disjoint indices (indices1, indices2) of length k,
    where every index is between 0 and n (exclusive).

    Examples
    --------
    >>> rng = np.random.default_rng(42)
    >>> indices_gen = SwapIndexGenerator(rng=rng, n=9)
    >>> indices_gen(2)
    (array([3, 0]), array([7, 2]))
    >>> indices_gen(2)
    (array([4, 6]), array([1, 5]))
    >>> indices_gen(1)
    (array([6]), array([7]))

    If the size is too large, the largest possible will be returned:
    >>> indices_gen(10)
    (array([7, 0, 4, 2]), array([3, 5, 1, 8]))
    """

    def __init__(self, rng, n: int):
        assert n >= 2
        self.rng = rng
        self.indices = np.arange(n)
        self.permutation = self.rng.permutation(self.indices)

    def __call__(self, size: int):
        assert size >= 1

        # Get 2 * size elements
        size = min(size, len(self.indices) // 2)
        chosen, self.permutation = (
            self.permutation[: 2 * size],
            self.permutation[2 * size :],
        )

        # Too few - generate a new permutation and try again
        if len(chosen) < 2 * size:
            self.permutation = self.rng.permutation(self.indices)
            return self.__call__(size=size)

        return chosen[:size], chosen[size:]


@dataclasses.dataclass(init=False, repr=True, eq=False)
class Permutation(Correlator):
    def __init__(
        self,
        *,
        weights=None,
        iterations=1000,
        tol=0.01,
        correlation_type="pearson",
        random_state=None,
        verbose=False,
    ):
        """Create a Permutation instance, which induces correlations
        between variables in X by randomly shuffling rows within each column.

        Parameters
        ----------
        weights : 2d numpy array or None, optional
            Elementwise weights for the target correlation matrix.
            The default is None, which corresponds to uniform weights.
        iterations : int, optional
            Maximal number of iterations to run. Each iterations consists of
            one loop over all variables. Choosing 0 means infinite iterations.
            The default is 1000.
        tol : float, optional
            Tolerance for stopping criteria. Will stop when
            norm(desired_corr - actual_corr) < tol. The default is 0.05.
        correlation_type : str, optional
            Either "pearson" or "spearman". The default is "pearson".
        random_state : np.random.Generator, int or None, optional
            A random state for the random number generator. The default is None.
        verbose : bool, optional
            Whether or not to print information. The default is False.

        Notes
        -----
        The paper "Correlation control in small-sample Monte Carlo type
        simulations I: A simulated annealing approach" by Vořechovský et al.
        proposes using simulated annealing. We implement a simple randomized
        hill climbing procedure instead, because it is good enough.
          - https://www.sciencedirect.com/science/article/pii/S0266892009000113
          - https://en.wikipedia.org/wiki/Hill_climbing

        Examples
        --------
        >>> rng = np.random.default_rng(42)
        >>> X = rng.normal(size=(100, 2))
        >>> float(sp.stats.pearsonr(*X.T).statistic)
        0.1573...
        >>> correlation_matrix = np.array([[1, 0.7], [0.7, 1]])
        >>> perm_trans = Permutation(random_state=0).set_target(correlation_matrix)
        >>> X_transformed = perm_trans(X)
        >>> float(sp.stats.pearsonr(*X_transformed.T).statistic)
        0.6832...

        For large matrices, it often makes sense to first use Iman-Conover
        to get a good initial solution, then give it to Permutation.
        Start by creating a large correlation matrix:

        >>> variables = 25
        >>> correlation_matrix = np.ones((variables, variables)) * 0.7
        >>> np.fill_diagonal(correlation_matrix, 1.0)
        >>> perm_trans = Permutation(iterations=250, tol=1e-6,
        ...                                    random_state=0, verbose=True)
        >>> perm_trans = perm_trans.set_target(correlation_matrix)

        Create data X, then transform using Iman-Conover:

        >>> X = rng.normal(size=(10 * variables, variables))
        >>> perm_trans._error(correlation_matrix, np.corrcoef(X, rowvar=False))
        0.4846...
        >>> ic_trans = ImanConover().set_target(correlation_matrix)
        >>> X_ic = ic_trans(X)
        >>> perm_trans._error(correlation_matrix, np.corrcoef(X_ic, rowvar=False)) # Error after Iman-Conover
        0.0071...
        >>> X_ic_pc = perm_trans(X_ic)
        Running permutation correlator for 250 iterations.
         Iter     25  Error: 0.007147 Swaps:  6
         Iter     50  Error: 0.007068 Swaps:  4
         Iter     75  Error: 0.006949 Swaps:  3
         Iter    100  Error: 0.006141 Swaps:  2
         Iter    125  Error: 0.005279 Swaps:  1
         Iter    150  Error: 0.004611 Swaps:  1
         Iter    175  Error: 0.004087 Swaps:  1
         Iter    200  Error: 0.003413 Swaps:  1
         Iter    225  Error: 0.003078 Swaps:  1
         Iter    250  Error: 0.002702 Swaps:  1
        >>> # Error after Iman-Conover + permutation
        >>> perm_trans._error(correlation_matrix, np.corrcoef(X_ic_pc, rowvar=False))
        0.0026...
        """

        if not (weights is None or np.all(weights > 0)):
            raise ValueError("`weights` must have positive entries.")
        if not (isinstance(iterations, int) and iterations >= 0):
            raise ValueError("`iterations` must be non-negative integer.")
        if not isinstance(tol, float) and tol > 0:
            raise ValueError("`tol` must be a positive float.")
        if not isinstance(verbose, bool):
            raise TypeError("`verbose` must be boolean")

        self.iters = iterations
        self.tol = tol
        self.rng = np.random.default_rng(random_state)
        self.verbose = verbose
        self.correlation_type = correlation_type

    def set_target(self, correlation_matrix, *, weights=None):
        """Set the target correlation matrix.

        Parameters
        ----------
        correlation_matrix : np.ndarray
            The target correlation matrix, with shape (variables, variables).
        weights : np.ndarray or None, optional
            A weight matrix, with shape (variables, variables).
            The default is None, which corresponds to 1 in every entry.
        """
        super().set_target(correlation_matrix, cholesky=False)
        weights = np.ones_like(self.C) if weights is None else weights
        self.weights = weights / np.sum(weights)
        self.triu_indices = np.triu_indices(self.C.shape[0], k=1)
        return self

    def _error(self, observed, target):
        """Compute RMSE over upper triangular part of corr(X) - target."""
        idx = self.triu_indices  # Get upper triangular indices (ignore diag)
        weighted_residuals_sq = self.weights[idx] * (observed[idx] - target[idx]) ** 2.0
        return float(np.sqrt(np.sum(weighted_residuals_sq)))

    @staticmethod
    def subiters(n, i):
        """Number of sub-iterations (swaps) per iteration."""
        # Use longer swap lengths in early iterations. The last half
        # of the iterations will use 1 sub-iteration. The second half of the
        # first half will use 2, and so forth. The pattern is:
        # n = 2 => [2, 1]
        # n = 4 => [3, 2, 1, 1]
        # n = 4 => [4, 3, 2, 2, 1, 1, 1, 1]
        # n = 8 => [5, 4, 3, 3, 2, 2, 2, 2, 1, 1, 1, 1, 1, 1, 1, 1]
        # This function computes a closed-form solution to indexing such
        # a list (of length n) at index i.
        C = np.log2(n) + 1
        return int(np.ceil(C ** (1 - (2 * i / n))))

    def __call__(self, X):
        """Cycle through through columns (variables), and for each
        column it swaps random rows (observations). If the result
        leads to a smaller error (correlation closer to target), then it is
        kept. If not we try again.

        Parameters
        ----------
        X : np.ndarray
            A matrix with shape (observations, variables).

        Returns
        -------
        A copy of X where rows within each column are shuffled.
        """
        self._validate_X(X, check_rows_cols=False)  # Allow rows <= columns

        num_obs, num_vars = X.shape
        if not (isinstance(X, np.ndarray) and X.ndim == 2):
            raise ValueError("`X` must be a 2D numpy array.")
        if not num_vars == self.C.shape[0]:
            raise ValueError(
                "Number of variables in `X` does not match `correlation_matrix`."
            )

        if self.verbose:
            print(
                f"Running permutation correlator for {self.iters if self.iters else 'inf'} iterations."
            )

        def product(iterations_gen, variables_gen):
            # itertools.product only works for finite inputs, so we need this
            for i in iterations_gen:
                for j in variables_gen:
                    yield (i, j)

        # Set up loop generator
        iter_gen = range(1, self.iters + 1) if self.iters else itertools.count(1)
        loop_gen = product(iter_gen, range(num_vars))  # (iteration, k)
        swaps_gen = SwapIndexGenerator(rng=self.rng, n=num_obs)

        # Set up variables that are tracked in the main loop
        corr_mat = CorrelationMatrix(
            X, correlation_type=self.correlation_type, check=False
        )
        current_error = self._error(observed=corr_mat[:, :], target=self.C)

        # Main loop. For each iteration, k cycles through all variables.
        # This parametrizes the algorithm so iterations is less sensitive to k.
        for iteration, k in loop_gen:
            print_iter = iteration % max(self.iters // 10, 1) if self.iters else 1000
            num_swaps = self.subiters(
                n=self.iters if self.iters else 10_000, i=iteration
            )
            if self.verbose and print_iter == 0 and k == 0:
                print(
                    f" Iter {iteration:>6}  Error: {current_error:.6f} Swaps: {num_swaps:>2}"
                )

            # Create two disjoint sets of swaps, e.g. [1, 3] and [4, 8]
            i, j = swaps_gen(num_swaps)

            # Attempt to swap
            new_corr_col = corr_mat.update_column(col=k, i=i, j=j)
            old_corr_col = corr_mat[k, :]  # Col or row is arbitrary
            target_corr_col = self.C[k, :]  # Col or row is arbitrary
            w = self.weights[k, :]
            old_error = np.average((target_corr_col - old_corr_col) ** 2, weights=w)
            new_error = np.average((target_corr_col - new_corr_col) ** 2, weights=w)

            # Better solution was found, store the change
            if new_error < old_error:
                corr_mat.commit(col=k, i=i, j=j)

            # Check convergence on every cycle through columns (k==0)
            if k == 0:
                current_error = self._error(corr_mat[:, :], self.C)
                if current_error < self.tol:
                    if self.verbose:
                        print(
                            f""" Terminating at iteration {iteration} due to tolerance. Error: {current_error:.6f}"""
                        )
                    return corr_mat.X

        return corr_mat.X  # The permuted data stored in CorrelationMatrix


def decorrelate(X, remove_variance=True):
    """Removes correlations or covariance from data X.

    Examples
    --------
    >>> X = np.array([[1. , 1. ],
    ...               [2. , 1.1],
    ...               [2.1, 3. ]])
    >>> X_decorr = decorrelate(X)

    The result has covariance equal to the identity matrix:

    >>> np.cov(X_decorr, rowvar=False).round(6)
    array([[1., 0.],
           [0., 1.]])

    The mean is preserved:

    >>> np.allclose(np.mean(X, axis=0), np.mean(X_decorr, axis=0))
    True

    The variance is removed:

    >>> np.var(X, axis=0)
    array([0.24666667, 0.84666667])
    >>> np.var(X_decorr, axis=0, ddof=1)
    array([1., 1.])

    We can optionally decorrelate while preserving the variance:

    >>> X_decorr = decorrelate(X, remove_variance=False)
    >>> np.cov(X_decorr, rowvar=False).round(6)
    array([[0.246667, 0.      ],
           [0.      , 0.846667]])
    >>> np.allclose(np.mean(X, axis=0), np.mean(X_decorr, axis=0))
    True
    """
    mean = np.mean(X, axis=0)
    var = np.var(X, axis=0, ddof=0)
    cov = np.cov(X, rowvar=False)

    L = np.linalg.cholesky(cov)  # L @ L.T = cov
    if not remove_variance:
        L = L / np.sqrt(var)

    # Computes X = (X - mean) @ inv(L).T
    X = sp.linalg.solve_triangular(L, (X - mean).T, lower=True).T

    return mean + X


class CorrelationMatrix:
    """Facilitates fast updates of a correlation matrices when swapping indices
    in the original data set X.

    The naive approach is to re-compute the entire correlation matrix after a
    swap, which has cost O(m n^2) if the data has shape (m, n). Some time can
    be saved if we realize that if we swap indices if in column k, then only
    the k'th row and column in the correlation matrix changes - this brings
    the cost down to O(m n).

    The final realization, which we implement here, exploits the fact that

      corr(x, y) = sum (x - E[x]) (y - E[y]) / (std(x) * std(y))
                 = sum (xy - x E[y] - y E[x] + E[x] E[y] ) / (std(x) * std(y))

    is mostly unchanged when indices are swapped. In fact, every term above
    except the first term xy is unchanged by swaps. The only thing we need
    to compute is the change in sum_i x_i * y_i. If `s` indices are swapped,
    then this can be done in O(s n) time. Since `s` is small this is more or
    less O(n) time - a vast improvement over the naive O(m n^2) approach.
    In practice vectorization matters too, but this approach is orders of
    magnitude faster on large data sets.

    Examples
    --------
    >>> rng = np.random.default_rng(42)
    >>> X = rng.normal(size=(9, 4))
    >>> computation = CorrelationMatrix(X)
    >>> computation[:].round(3)  # Access the array and print it
    array([[ 1.   ,  0.408,  0.618,  0.063],
           [ 0.408,  1.   ,  0.318, -0.6  ],
           [ 0.618,  0.318,  1.   , -0.151],
           [ 0.063, -0.6  , -0.151,  1.   ]])

    A single swap:

    >>> computation.update_column(col=0, i=2, j=3)
    array([1.        , 0.37191405, 0.62817264, 0.09671987])
    >>> X[2, 0], X[3, 0] = X[3, 0], X[2, 0]

    Verify that the column/row is correct:

    >>> np.corrcoef(X, rowvar=False)[:, 0]
    array([1.        , 0.37191405, 0.62817264, 0.09671987])
    >>> X[2, 0], X[3, 0] = X[3, 0], X[2, 0]

    A series of swaps:

    >>> computation[:].round(3)
    array([[ 1.   ,  0.408,  0.618,  0.063],
           [ 0.408,  1.   ,  0.318, -0.6  ],
           [ 0.618,  0.318,  1.   , -0.151],
           [ 0.063, -0.6  , -0.151,  1.   ]])
    >>> computation.update_column(col=0, i=[0, 1], j=[2, 3])
    array([ 1.        , -0.64630365,  0.42642021,  0.32491853])
    >>> X[[0, 1], 0], X[[2, 3], 0] = X[[2, 3], 0], X[[0, 1], 0]
    >>> np.corrcoef(X, rowvar=False).round(3)
    array([[ 1.   , -0.646,  0.426,  0.325],
           [-0.646,  1.   ,  0.318, -0.6  ],
           [ 0.426,  0.318,  1.   , -0.151],
           [ 0.325, -0.6  , -0.151,  1.   ]])
    """

    def __init__(self, X, correlation_type="pearson", check=True):
        valid_corrs = ("pearson", "spearman")
        assert correlation_type in valid_corrs
        assert X.ndim == 2

        # Correlation type
        self.correlation_type = correlation_type
        self.check = check

        # Store a copy of the data in the original space
        self.X = X.copy()

        # Store a copy of the data in the space we're inducing correlations in
        if self.correlation_type == "pearson":
            self.X_ = self.X
        elif self.correlation_type == "spearman":
            # Spearman(X) = Pearson(rank(X))
            self.X_ = np.apply_along_axis(sp.stats.rankdata, axis=0, arr=self.X)
        else:
            raise ValueError(
                f"`correlation_type` must be in {valid_corrs}, got {correlation_type}"
            )

        # Compute the correlation matrix of the data
        self.m, self.n = self.X_.shape
        X_centered = self.X_ - np.mean(self.X_, axis=0)
        self.numerator = (X_centered.T @ X_centered) / self.m
        self.denominator = np.std(X_centered, axis=0)
        if np.any(np.isclose(self.denominator, 0)):
            raise ValueError("X has one or several constant columns")

        self.corr_mat = (self.numerator / self.denominator[None, :]) / self.denominator[
            :, None
        ]

    def __repr__(self):
        return repr(self.corr_mat)

    def __getitem__(self, *args, **kwargs):
        return self.corr_mat.__getitem__(*args, **kwargs)

    def commit(self, col, i, j):
        """Commit a swap, storing new data and new correlation matrix."""

        # Compute everything we need to update internal state
        delta_numerator = self._delta_numerator(col, i, j)
        delta_column = delta_numerator / (
            self.m * self.denominator * self.denominator[col]
        )

        # Update correlation and the denominator
        self.corr_mat[:, col] += delta_column
        self.corr_mat[col, :] += delta_column
        self.numerator[:, col] += delta_numerator
        self.numerator[col, :] += delta_numerator

        # Update data
        self.X_[i, col], self.X_[j, col] = self.X_[j, col], self.X_[i, col]
        if self.correlation_type == "spearman":  # Update original too
            self.X[i, col], self.X[j, col] = self.X[j, col], self.X[i, col]
        return self

    def _delta_numerator(self, col, i, j):
        """Compute the delta in the numerator when swapping."""
        if self.check:
            assert isinstance(col, int)
            assert 0 <= col < self.n
            if isinstance(i, int):
                i = [i]
            if isinstance(j, int):
                j = [j]

            assert len(i) == len(j)

            if set(i).intersection(set(j)):
                raise ValueError(f"Swaps must be two disjoint sets, got {i} and {j}")

        # Vectorized over all swaps
        row_i = self.X_[i, :]
        row_j = self.X_[j, :]
        entry_ic = row_i[:, col]
        entry_jc = row_j[:, col]

        delta_numerator = np.sum(
            (row_i - row_j) * (entry_jc - entry_ic)[:, None], axis=0
        )
        delta_numerator[col] = 0.0
        return delta_numerator

    def delta_column(self, col, i, j):
        """Returns the change in the column `col` in the correlation matrix
        when rows i and j are swapped. To save a change, use `.commit()`."""

        diff = self._delta_numerator(col, i, j)
        return diff / (self.m * self.denominator * self.denominator[col])

    def update_column(self, col, i, j):
        """Returns the new value of column `col` in the correlation matrix
        when rows i and j are swapped. To save a change, use `.commit()`."""

        delta = self.delta_column(col, i, j)
        return self.corr_mat[:, col] + delta


@dataclasses.dataclass(init=False, repr=True, eq=False)
class Composite(Correlator):
    """A composition where we first run ImanConover, then Permutation."""

    def __init__(self, *args, **kwargs):
        self.iman_conover_correlator = ImanConover()
        self.permutation_correlator = Permutation(*args, **kwargs)

    def set_target(self, correlation_matrix, *, weights=None):
        self.iman_conover_correlator.set_target(correlation_matrix)
        self.permutation_correlator.set_target(correlation_matrix, weights=weights)
        return self

    def __call__(self, X):
        # First run ImanConover to get a good starting point
        X_ic = self.iman_conover_correlator(X)
        # Then run Permutation
        return self.permutation_correlator(X_ic)


if __name__ == "__main__":
    import pytest
    import matplotlib.pyplot as plt

    pytest.main(args=[__file__, "--doctest-modules", "-v", "--capture=sys"])

    if True:
        rng = np.random.default_rng(2)
        p = 10
        n = 999

        sampler = sp.stats.qmc.Halton(d=p, seed=42, scramble=False)
        samples = sampler.random(n=n)
        samples = np.vstack(
            [rng.permutation(np.linspace(0 + 1e-6, 1 - 1e-6, num=n)) for k in range(p)]
        ).T

        X = np.zeros_like(samples)
        for j in range(X.shape[1]):
            func_i = int(rng.integers(0, 4))
            func = [
                sp.stats.uniform(),
                sp.stats.expon(),
                sp.stats.norm(),
                sp.stats.lognorm(s=1),
            ]
            func = func[func_i]
            # func = sp.stats.uniform()
            X[:, j] = func.ppf(samples[:, j]) * rng.normal(loc=5, scale=2)

        plt.figure(figsize=(4, 4))
        plt.title("Original data set")
        plt.scatter(X[:, 0], X[:, 1], s=1)
        plt.show()

        target = np.ones((p, p)) * 0.5
        np.fill_diagonal(target, 1.0)

        # Create permutation correlator
        correlator = Permutation(verbose=True, tol=1e-4, iterations=10_000)
        correlator.set_target(target)

        # First try cholesky and
        X_chol = Cholesky().set_target(target)(X)
        corr_X_chol = np.corrcoef(X_chol, rowvar=False)
        error = correlator._error(corr_X_chol, target)
        print(f"Cholesky error: {error:.4f}")
        plt.figure(figsize=(4, 4))
        plt.title(f"Cholesky error: {error:.4f}")
        plt.scatter(X_chol[:, 0], X_chol[:, 1], s=1)
        plt.show()

        # First try cholesky and
        X_ic = ImanConover().set_target(target)(X)
        corr_X_ic = np.corrcoef(X_ic, rowvar=False)
        error = correlator._error(corr_X_ic, target)
        print(f"ImanConover error: {error:.4f}")
        plt.figure(figsize=(4, 4))
        plt.title(f"ImanConover error: {error:.4f}")
        plt.scatter(X_ic[:, 0], X_ic[:, 1], s=1)
        plt.show()

        X_pc = correlator(X)
        corr_X_pc = np.corrcoef(X_pc, rowvar=False)
        error = correlator._error(corr_X_pc, target)
        print(f"Permutation error: {error:.4f}")
        plt.figure(figsize=(4, 4))
        plt.title(f"Permutation error: {error:.4f}")
        plt.scatter(X_pc[:, 0], X_pc[:, 1], s=1)
        plt.show()



import numpy as np
import pytest
from probabilit.correlation import nearest_correlation_matrix


class TestNearestCorrelationMatrix:
    @pytest.mark.parametrize("variables", range(2, 100, 10))
    def test_nearest_correlation_matrix(self, variables):
        """Test that we can cholesky decompose the solution."""

        rng = np.random.default_rng(variables)

        # Create a correlation matrix
        observations = rng.normal(size=(variables * 2, variables))
        matrix = np.corrcoef(observations, rowvar=False)

        # Taking the cholesky decomposition should work just fine
        np.linalg.cholesky(matrix)

        # Mess it up
        matrix = matrix + rng.normal(size=matrix.shape, scale=0.1)
        matrix = matrix - np.identity(variables) * np.mean(np.diag(matrix))

        # Now the cholesky decomposition should fail
        with pytest.raises(np.linalg.LinAlgError):
            np.linalg.cholesky(matrix)

        # Adjust the matrix to its nearest correlation matrix
        correlation_matrix = nearest_correlation_matrix(matrix)

        # Taking the cholesky decomposition should work now
        np.linalg.cholesky(correlation_matrix)

        # Diagonal entries should be 1.0 and the matrix should be symmetric
        assert np.allclose(np.diag(correlation_matrix), 1.0)
        assert np.allclose(correlation_matrix, correlation_matrix.T)

    def test_nearest_correlation_matrix_on_matlab_example(self):
        """These matrices are from the 'nearcorr' docs:
        https://www.mathworks.com/help/stats/nearcorr.html
        """
        # The matrix we want to adjust to become a correlation matrix
        A = np.array(
            [
                [1.0, 0.0, 0.0, 0.0, -0.936],
                [0.0, 1.0, -0.55, -0.3645, -0.53],
                [0.0, -0.55, 1.0, -0.0351, 0.0875],
                [0.0, -0.3645, -0.0351, 1.0, 0.4557],
                [-0.936, -0.53, 0.0875, 0.4557, 1.0],
            ]
        )

        W = np.array(
            [
                [0.0, 1.0, 0.1, 0.15, 0.25],
                [1.0, 0.0, 0.05, 0.025, 0.15],
                [0.1, 0.05, 0.0, 0.25, 1.0],
                [0.15, 0.025, 0.25, 0.0, 0.25],
                [0.25, 0.15, 1.0, 0.25, 0.0],
            ]
        )

        matlab_Y = np.array(
            [
                [1.0, 0.0014, 0.0287, -0.0222, -0.8777],
                [0.0014, 1.0, -0.498, -0.7268, -0.4567],
                [0.0287, -0.498, 1.0, -0.0358, 0.0878],
                [-0.0222, -0.7268, -0.0358, 1.0, 0.4465],
                [-0.8777, -0.4567, 0.0878, 0.4465, 1.0],
            ]
        )

        # The smallest eigenvalue of A is -0.1244...
        # The smallest eigenvalue of Y is 1.088e-06
        Y = nearest_correlation_matrix(A, weights=W)

        # Matlab output has 4 digits, so atol is set to 1e-4 here
        assert np.allclose(Y, matlab_Y, atol=1e-4)


if __name__ == "__main__":
    import pytest

    pytest.main(args=[__file__, "--doctest-modules", "-v", "-l"])



from probabilit.modeling import (
    EmpiricalDistribution,
    Constant,
    Log,
    Exp,
    Distribution,
    Floor,
    Equal,
    All,
    Min,
    Max,
)
from probabilit.distributions import Triangular
import numpy as np


class TestModelingExamples:
    def test_die_problem(self):
        """If we throw 2 die, what is the probability that each one ends up
        with the same number?"""

        die1 = Floor(1 + Distribution("uniform") * 6)
        die2 = Floor(1 + Distribution("uniform") * 6)
        equal = Equal(die1, die2)

        samples = equal.sample(999, random_state=42)

        np.testing.assert_allclose(samples.mean(), 1 / 6, atol=0.001)

    def test_estimating_pi(self):
        """Consider the unit square [0, 1]^2. The area of the square is 1.
        The area of a quarter circle is pi * r^2 / 4 = pi / 4.
        So the fraction (quarter circle area) / (total area) = pi / 4.

        Use this to estimate pi.
        """

        x = Distribution("uniform")
        y = Distribution("uniform")
        inside = x**2 + y**2 < 1
        pi_estimate = 4 * inside

        samples = pi_estimate.sample(9999, random_state=42)
        np.testing.assert_allclose(samples.mean(), np.pi, atol=0.01)

    def test_broken_stick_problem(self):
        """Consider a stick of length 1. Pick two points uniformly at random on
        the stick, and break the stick at those points. What is the probability
        that the three segments obtained in this way form a triangle?

        Of course this is the probability that no one of the short sticks is
        longer than 1/2. This probability turns out to be 1/4.

        https://sites.math.duke.edu/education/webfeatsII/gdrive/Team%20D/project/brokenstick.htm
        https://mathoverflow.net/questions/2014/if-you-break-a-stick-at-two-points-chosen-uniformly-the-probability-the-three-r
        """

        # Cuts along the stick
        cut1 = Distribution("uniform", loc=0, scale=1)
        cut2 = Distribution("uniform", loc=0, scale=1)

        # The lengths
        length1 = Min(cut1, cut2)
        length2 = Max(cut1, cut2) - Min(cut1, cut2)
        length3 = 1 - Max(cut1, cut2)

        # No one of the short sticks is longer than 1/2 <=> all are shorter
        prob = All(length1 < 1 / 2, length2 < 1 / 2, length3 < 1 / 2)

        samples = prob.sample(9999, random_state=42)
        np.testing.assert_allclose(samples.mean(), 1 / 4, atol=0.01)

    def test_mutual_fund_problem(self):
        """Suppose you save 1200 units of money per year and that the yearly
        interest rate has a   distribution `N(1.11, 0.15)`.
        How much money will you have over a 20 year horizon?

        From: https://curvo.eu/backtest/en/market-index/sp-500?currency=eur
        In the last 33 years, the S&P 500 index (in EUR) had a compound annual
        growth rate of 10.83%, a standard deviation of 15.32%, and a Sharpe ratio of 0.66.
        """

        saved_per_year = 1200
        returns = 0
        for year in range(20):
            interest = Distribution("norm", loc=1.11, scale=0.15)
            returns = returns * interest + saved_per_year
        samples = returns.sample(999, random_state=42)

        # Regression test essentially
        np.testing.assert_allclose(samples.mean(), 76583.58738496085)
        np.testing.assert_allclose(samples.std(), 33483.2245611436)

    def test_total_person_hours(self):
        """Based on Example 19.2 from Risk Analysis: A Quantitative Guide, 3rd Edition by David Vose.

        Estimate the number of person-hours requires to rivet 562 plates of a ship's hull.
        The quickest anyone has ever riveted a single plate is 3h 45min, while the worst time recorded is 5h 30min.
        Most likely value is estimated to be 4h 15min.
        What is the total person-hours?

        Naively, we could model the problem as:
        total_person_hours = 562 * Triangular(3.75, 4.25, 5.5),
        but note that the triangular distribution here models the uncertainty of an individual plate,
        but we are using it as if it were the distribution of the average time for 562 plates.

        A straight forward approach that gives the correct answer is to add 562 triangular distributions.
        """

        rng = np.random.default_rng(42)
        num_rivets = 562
        total_person_hours = 0

        for i in range(num_rivets):
            total_person_hours += Triangular(
                low=3.75, mode=4.25, high=5.5, low_perc=0, high_perc=1.0
            )

        num_samples = 1000
        res_total_person_hours = total_person_hours.sample(
            num_samples, random_state=rng
        )

        # The mean and standard deviation of a Triangular(3.75, 4.25, 5.5) are 4.5 and 0.368,
        # so by the Central Limit Theoreom we have that
        # total_person_hours = Normal(4.5 * 562, 0.368 * sqrt(562)) = Normal(2529, 8.724)
        expected_mean = 4.5 * num_rivets
        expected_std = 0.368 * np.sqrt(num_rivets)

        sample_mean = np.mean(res_total_person_hours)
        sample_std = np.std(res_total_person_hours, ddof=1)

        # Within 2% of theoretical values
        np.testing.assert_allclose(sample_mean, expected_mean, rtol=0.02)
        np.testing.assert_allclose(sample_std, expected_std, rtol=0.02)

    def test_conditional_if_statement(self):
        """Suppose mens height has distribution N(176, 7.1).
        What is the distribution of the difference between height of two men?
        Caveat: there is a 10% chance that the two men are identical twins,
        and in that case their height should be perfectly equal.
        """

        # Height of two random men
        height1 = Distribution("norm", loc=176, scale=7.1)
        height2 = Distribution("norm", loc=176, scale=7.1)

        # If they are twins, their height should be perfectly correlated
        is_twin = Distribution("bernoulli", p=0.1)

        # height2 = IF(is_twin, height1, height2)
        height2 = is_twin * height1 + (1 - is_twin) * height2

        # This is the answer to the question
        (abs(height2 - height1)).sample(999, random_state=42)

        # At least one of the realizations should be identical
        assert np.any(np.isclose(height1.samples_, height2.samples_))

    def test_fault_controlled_owc_correlation(self):
        """
        Test that oil-water contact (OWC) correlation between segments
        depends on fault state in geological modeling.

        When fault is open (leaking): Seg2 should have same contact as Seg1
        When fault is closed: Seg2 should follow independent distribution (1950-2000m)
        """
        # Setup
        rng = np.random.default_rng(42)
        n_samples = 100

        # Seg1: OWC = 2000 +/- 5 m (observed segment)
        owc1 = Distribution("uniform", loc=1995, scale=10)

        # Fault state: 30% probability of being open (leaking)
        fault_is_open = Distribution("bernoulli", p=0.3)

        # Seg2: Conditional OWC based on fault state
        # If fault open: same as Seg1
        # If fault closed: independent uniform distribution 1950-2000m
        owc2 = fault_is_open * owc1 + (1 - fault_is_open) * Distribution(
            "uniform", loc=1950, scale=50
        )

        # Generate samples
        owc2_samples = owc2.sample(n_samples, random_state=rng)

        # Get individual component samples for verification
        owc1_samples = owc1.samples_
        fault_samples = fault_is_open.samples_.astype(bool)
        owc2_samples = owc2.samples_

        # Verify fault-controlled correlation
        for i in range(n_samples):
            if fault_samples[i]:  # Fault is open (leaking)
                assert np.isclose(owc2_samples[i], owc1_samples[i], rtol=1e-10), (
                    f"Sample {i}: When fault is open, Seg2 OWC ({owc2_samples[i]:.2f}) "
                    f"should equal Seg1 OWC ({owc1_samples[i]:.2f})"
                )
            else:  # Fault is closed
                assert 1950 <= owc2_samples[i] <= 2000, (
                    f"Sample {i}: When fault is closed, Seg2 OWC ({owc2_samples[i]:.2f}) "
                    f"should be in independent range [1950-2000m]"
                )

        # Additional statistical checks
        open_fault_count = np.sum(fault_samples)
        closed_fault_count = n_samples - open_fault_count

        # Verify we have reasonable sample sizes for both scenarios
        assert open_fault_count > 0, "Should have some samples with open fault"
        assert closed_fault_count > 0, "Should have some samples with closed fault"

    def test_stopping_distance_problem(self):
        """A car is driving at a speed of 100 km / hour. Suddenly it has to slam
        the breaks. What is the breaking distance of the car?

        The main uncertainty is the speed and the coefficient of friction.
        For fun we can also account for some variation in the gravitational
        constant g, since it depends on altitude and where on earth we are."""

        # The kinetic energy is E = 1/2 m v^2
        # The breaking for has to stop this energy
        # Energy = force x distance
        # 1/2 m v^2 = (mu g m) x distance
        # distance = v^2 / (2 mu g)

        # Around 3% relative error, but on multiplicative scale
        velocity_kmh = Exp(Distribution("norm", 0, 0.03)) * 100
        velocity_ms = velocity_kmh / 3.6  # Convert to meters per second
        g = Distribution("norm", loc=9.8220, scale=0.0020)

        # On dry concrete/asphalt, around 0.7 is reasonable
        # https://en.wikibooks.org/wiki/Physics_Study_Guide/Frictional_coefficients
        mu_dry = Distribution("norm", loc=0.7, scale=0.02)
        distance_dry = velocity_ms**2 / (2 * mu_dry * g)

        # Sample and create a simple regression/snapshot test
        samples = distance_dry.sample(999, random_state=42, method="lhs")
        np.testing.assert_allclose(np.mean(samples), 56.258925)
        np.testing.assert_allclose(np.std(samples), 3.707973)

        # On wet concrete mu=0.58, but on wet asphalt mu = 0.53.
        # Here we model a 50/50 chance of being on either (mixture distribution)
        # and a sigma of 0.02 either way.
        is_concrete = Distribution("bernoulli", p=0.5)
        mu_wet = Distribution("norm", loc=0.53, scale=0.02) + 0.05 * is_concrete
        distance_wet = velocity_ms**2 / (2 * mu_wet * g)

        # Sample and create a simple regression/snapshot test
        samples = distance_wet.sample(999, random_state=42, method="lhs")
        np.testing.assert_allclose(np.mean(samples), 71.13491)
        np.testing.assert_allclose(np.std(samples), 5.830891)


def test_copying():
    # Create a graph
    mu = Distribution("norm", loc=0, scale=1)
    a = Distribution("norm", loc=mu, scale=Constant(0.5))

    # Create a copy
    a2 = a.copy()

    # The copy is not the same object
    assert a2 is not a

    # However, the IDs match and they are equal
    assert a2 == a and (a2._id == a._id)

    # The same holds for parents - they are copied
    assert a2.kwargs["loc"] is not a.kwargs["loc"]

    a.sample()
    assert hasattr(a, "samples_")
    assert not hasattr(a2, "samples_")

    # Now create a copy and ensure samples are copied too
    a3 = a.copy()
    assert hasattr(a3, "samples_")
    assert a3.samples_ is not a.samples_


def test_constant_arithmetic():
    # Test that converstion with int works
    two = Constant(2)
    result = two + 2
    np.testing.assert_allclose(result.sample(), 4)

    # Test that subtraction works both ways
    two = Constant(2)
    five = Constant(5)
    result1 = five - two
    result2 = 5 - two
    result3 = five - two
    np.testing.assert_allclose(result1.sample(), result2.sample())
    np.testing.assert_allclose(result1.sample(), result2.sample())
    np.testing.assert_allclose(result1.sample(), result3.sample())
    np.testing.assert_allclose(result1.sample(), 5 - 2)

    # Test that divison works both ways
    two = Constant(2)
    five = Constant(5)
    result1 = five / two
    result2 = 5 / two
    result3 = five / two
    np.testing.assert_allclose(result1.sample(), result2.sample())
    np.testing.assert_allclose(result1.sample(), result2.sample())
    np.testing.assert_allclose(result1.sample(), result3.sample())
    np.testing.assert_allclose(result1.sample(), 5 / 2)

    # Test absolute value and negation
    result = abs(-two)
    np.testing.assert_allclose(result.sample(), 2)

    # Test powers
    result = five**two
    np.testing.assert_allclose(result.sample(), 5**2)


def test_constant_expressions():
    # Test a few longer expressions
    two = Constant(2)
    five = Constant(5)
    result = two + two - five**2 + abs(-five)
    np.testing.assert_allclose(result.sample(), 2 + 2 - 5**2 + abs(-5))

    result = two / five - two**3 + Exp(5)
    np.testing.assert_allclose(result.sample(), 2 / 5 - 2**3 + np.exp(5))

    result = 1 / five - (Log(5) + Exp(Log(10)))
    np.testing.assert_allclose(result.sample(), 1 / 5 - (np.log(5) + 10))


def test_single_expression():
    # A graph with a single node is an edge-case
    samples = Constant(2).sample()
    np.testing.assert_allclose(samples, 2)


def test_constant_idempotent():
    for a in [-1, 0.0, 1.3, 3]:
        assert Constant(Constant(a)).value == Constant(a).value


def test_that_an_empirical_distribution_can_be_a_parameter():
    location = EmpiricalDistribution(data=[1, 2, 3, 3, 3, 3])
    result = Distribution("norm", loc=location, scale=1)
    (result**2).sample(99, random_state=42)


def test_that_distribution_params_with_transforms():
    # Plain old numbers work as arguments without raising any errors
    loc = 2
    samples1 = Distribution("norm", loc=loc).sample(99, random_state=0)

    # The same number wrapped in constant
    loc = Constant(2)
    samples2 = Distribution("norm", loc=loc).sample(99, random_state=0)

    # A more complex expression: loc = 0 + sqrt(9) - Log(2) = 0 + 3 - 1 = 2
    loc = Constant(0) + (Constant(9) ** 0.5) - Log(2.718281828459045)
    samples3 = Distribution("norm", loc=loc).sample(99, random_state=0)

    np.testing.assert_allclose(samples1, samples2)
    np.testing.assert_allclose(samples1, samples3)


def test_correlations():
    # Test that inducing correlations leads to samples with approximately the
    # same observed correlations as the desired correlations
    a = Distribution("norm", loc=0, scale=1)
    b = Distribution("norm", loc=0, scale=1)
    c = Distribution("norm", loc=0, scale=1)
    d = Distribution("norm", loc=0, scale=1)
    expression = a + b + c + d

    # This is not a valid correlation matrix (not pos.def) - but probabilit will fix that
    corr_mat = np.array([[1.0, 0.8, 0.5], [0.8, 1.0, 0.8], [0.5, 0.8, 1.0]])
    expression.correlate(a, b, c, corr_mat=corr_mat)

    expression.sample(999, random_state=42)

    observed_corr_mat = np.corrcoef(
        np.vstack([a.samples_, b.samples_, c.samples_, d.samples_])
    )
    desired_corr_mat = np.eye(4)
    desired_corr_mat[:3, :3] = corr_mat

    np.testing.assert_allclose(observed_corr_mat, desired_corr_mat, atol=0.075)


def test_all_correlations_at_unity():
    # 1. Test that a correlation matrices with all entries at 1.0 works
    # 2. Test that the resulting observed matrix is very close

    a = Distribution("norm", loc=0, scale=1)
    b = Distribution("norm", loc=0, scale=1)
    c = Distribution("norm", loc=0, scale=1)
    expression = a + b + c

    corr_mat = np.ones((3, 3))
    expression.correlate(a, b, c, corr_mat=corr_mat)
    expression.sample(999, random_state=42, method="lhs")

    obs_corr = np.corrcoef(np.array([a.samples_, b.samples_, c.samples_]))
    assert np.linalg.norm(obs_corr - corr_mat) <= 0.00015


if __name__ == "__main__":
    import pytest

    pytest.main(args=[__file__, "-v", "--capture=sys"])



"""
Probabilit uses distributions implemented in SciPy by default, e.g.:

>>> normal = Distribution("norm", loc=0, scale=1)
>>> gamma = Distribution("gamma", a=1)
>>> generalized_pareto = Distribution("genpareto", c=2)

For a full list, see:

  - https://docs.scipy.org/doc/scipy/reference/stats.html#continuous-distributions

In some cases it might make sense to implement our own custom distributions.
CUSTOM DISTRIBUTIONS SHOULD BE IMPLEMENTED SPARINGLY. WE DO NOT WANT TO GO DOWN
THE PATH OF RE-IMPLEMENTING SCIPY. For better or worse, scipy is a de-facto
standard and many are used to it. Using scipy also delegates documentation burden.

Some reasons to to this:

  1. Better naming and easier API, e.g. `Normal(...)` vs `Distribution("norm", ...)`
  2. Alternative parametrizations (typical example is the LogNorm)
  3. Distributions not found in SciPy

Most of the custom distributions are syntactic sugar:

>>> Distribution("uniform", loc=-1, scale=2)
Distribution("uniform", loc=-1, scale=2)
>>> Uniform(min=-1, max=1)
Distribution("uniform", loc=-1, scale=2)

"""

import numpy as np
import warnings
import scipy as sp
from probabilit.modeling import Distribution, Log, Exp, Sign


def Uniform(min=0, max=1):
    """Uniform distribution on [min, max)."""
    return Distribution("uniform", loc=min, scale=max - min)


def Normal(loc, scale):
    """Normal distribution parametrized by mean (loc) and std (scale)."""
    return Distribution("norm", loc=loc, scale=scale)


def TruncatedNormal(loc, scale, low=-np.inf, high=np.inf):
    """A truncated Normal distribution parametrized by mean (loc) and
    std (scale) defined on [low, high).

    Examples
    --------
    >>> distr = TruncatedNormal(loc=0, scale=1, low=3, high=3.3)
    >>> distr.sample(7, random_state=0).round(3)
    array([3.13 , 3.182, 3.146, 3.129, 3.095, 3.159, 3.099])
    """
    # (a, b) are defined in terms of loc and scale, so transform them
    a, b = (low - loc) / scale, (high - loc) / scale
    return Distribution("truncnorm", a=a, b=b, loc=loc, scale=scale)


class Lognormal(Distribution):
    def __init__(self, mean, std):
        """
        A Lognormal distribution with mean and std corresponding directly
        to the expected value and standard deviation of the resulting lognormal.

        Examples
        --------
        >>> samples = Lognormal(mean=2, std=1).sample(999, random_state=0)
        >>> float(np.mean(samples))
        2.00173...
        >>> float(np.std(samples))
        1.02675...

        Composite distributions work too:

        >>> mean = Distribution("expon", scale=1)
        >>> Lognormal(mean=mean, std=1).sample(5, random_state=0)
        array([0.86196529, 0.69165866, 0.41782557, 1.23340656, 2.90778578])
        """
        # Transform parameters (they can be numbers, distributions, etc)
        variance = Sign(std) * std**2  # Square it but keep the sign (so negative fails)
        sigma_squared = Log(1 + variance / (mean**2))
        sigma = (sigma_squared) ** (1 / 2)
        mu = Log(mean) - sigma_squared / 2

        # Call the parent class
        super().__init__(distr="lognorm", s=sigma, scale=Exp(mu))

    @classmethod
    def from_log_params(cls, mu, sigma):
        """
        Create a lognormal distribution from log-space parameters.
        Parameters correspond to the mean and standard deviation of the
        underlying normal distribution (i.e., the parameters of log(X) where
        X is the lognormal random variable).

        Examples
        --------
        >>> mu = Distribution("norm")
        >>> Lognormal.from_log_params(mu=mu, sigma=1).sample(5, random_state=0)
        array([1.99625633, 1.45244764, 1.19926216, 2.94150961, 4.47459182])
        """
        return Distribution("lognorm", s=sigma, scale=Exp(mu))


def PERT(minimum, mode, maximum, gamma=4.0):
    """Returns a Beta distribution, parameterized by the PERT parameters.

    A high gamma value means a more concentrated distribution.

    Examples
    --------
    >>> PERT(0, 6, 10)
    Distribution("beta", a=3.4, b=2.6, loc=0, scale=10)
    >>> PERT(0, 6, 10, gamma=10)
    Distribution("beta", a=7.0, b=5.0, loc=0, scale=10)
    """
    # Based on Wikipedia and another implementation:
    # https://en.wikipedia.org/wiki/PERT_distribution
    # https://github.com/Calvinxc1/PertDist/blob/6577394265f57153441b5908147d94115b9edeed/pert/pert.py#L80
    a, b, loc, scale = _pert_to_beta(minimum, mode, maximum, gamma=gamma)
    return Distribution("beta", a=a, b=b, loc=loc, scale=scale)


def Triangular(low, mode, high, low_perc=0.1, high_perc=0.9):
    """Find optimal scipy parametrization given (low, mode, high) and
    return Distribution("triang", loc=..., scale=..., c=...).

    This distribution does *not* work with composite distributions.
    The arguments must be numbers, they cannot be other distributions.

    Examples
    --------
    >>> Triangular(low=1, mode=5, high=9)
    Distribution("triang", loc=-2.2360679774997894, scale=14.472135954999578, c=0.5000000000000001)
    >>> Triangular(low=1, mode=5, high=9, low_perc=0.25, high_perc=0.75)
    Distribution("triang", loc=-8.656854249492383, scale=27.313708498984766, c=0.5)
    >>> Triangular(low=1, mode=5, high=9, low_perc=0, high_perc=1)
    Distribution("triang", loc=1, scale=8, c=0.5)
    """
    # A few comments on fitting can be found here:
    # https://docs.analytica.com/index.php/Triangular10_50_90

    if not ((0 <= low_perc <= 1.0) and (0 <= high_perc <= 1.0)):
        raise ValueError("Percentiles must be between 0 and 1.")

    # No need to optimize if low and high are boundaries of distribution support
    if np.isclose(low_perc, 0.0) and np.isclose(high_perc, 1.0):
        loc, scale, c = low, high - low, (mode - low) / (high - low)

    else:
        # Optimize parameters
        loc, scale, c = _fit_triangular_distribution(
            low=low,
            mode=mode,
            high=high,
            low_perc=low_perc,
            high_perc=high_perc,
        )
    return Distribution("triang", loc=loc, scale=scale, c=c)


def _fit_triangular_distribution(low, mode, high, low_perc=0.10, high_perc=0.90):
    """Returns a tuple (loc, scale, c) to be used with scipy.

    Examples
    --------
    >>> _fit_triangular_distribution(3, 8, 10, low_perc=0.10, high_perc=0.90)
    (-0.207..., 12.53..., 0.65...)
    >>> _fit_triangular_distribution(3, 8, 10, low_perc=0.4, high_perc=0.6)
    (-27.63..., 65.82..., 0.54...)
    >>> _fit_triangular_distribution(3, 8, 10, low_perc=0, high_perc=1.0)
    (3.00..., 6.99..., 0.71...)
    """

    def triangular_cdf(x, a, b, mode):
        """Calculate CDF of triangular distribution at point x"""
        if x <= a:
            return x * 0
        if x >= b:
            return x * 0 + 1.0
        if x <= mode:
            return ((x - a) ** 2) / ((b - a) * (mode - a))
        else:
            return 1 - ((b - x) ** 2) / ((b - a) * (b - mode))

    def equations(params):
        """System of equations to solve for a and b"""
        a, b = params

        # Calculate CDFs at the given percentile values
        cdf_low = triangular_cdf(low, a, b, mode)
        cdf_high = triangular_cdf(high, a, b, mode)

        # Return the difference from target percentiles
        return (cdf_low - low_perc, cdf_high - high_perc)

    # Initial guesses for a and b, the lower and upper bounds for support
    a0 = low - abs(mode - low)
    b0 = high + abs(high - mode)

    # Solve the system of equations
    a, b = sp.optimize.fsolve(equations, (a0, b0))
    rmse = np.sqrt(np.sum(np.array(equations([a, b])) ** 2))
    if rmse > 1e-6:
        warnings.warn(f"Optimization of Triangular params has {rmse=}")

    # Calculate the relative position of the mode, return (loc, scale, c)
    c = (mode - a) / (b - a)
    return float(a), float(b - a), float(c)


def _pert_to_beta(minimum, mode, maximum, gamma=4.0):
    """Convert the PERT parametrization to a beta distribution.

    Returns (a, b, loc, scale).

    Examples
    --------
    >>> _pert_to_beta(0, 3/4, 1)
    (4.0, 2.0, 0, 1)
    >>> _pert_to_beta(0, 30/4, 10)
    (4.0, 2.0, 0, 10)
    >>> _pert_to_beta(0, 9, 10, gamma=6)
    (6.4, 1.6, 0, 10)
    """
    # https://en.wikipedia.org/wiki/PERT_distribution
    if not (minimum < mode < maximum):
        raise ValueError(f"Must have {minimum=} < {mode=} < {maximum=}")
    if gamma <= 0:
        raise ValueError(f"Gamma must be positive, got {gamma=}")

    # Determine location and scale
    loc = minimum
    scale = maximum - minimum

    # Determine a and b
    a = 1 + gamma * (mode - minimum) / scale
    b = 1 + gamma * (maximum - mode) / scale

    return (a, b, loc, scale)


if __name__ == "__main__":
    import pytest

    pytest.main(args=[__file__, "--doctest-modules", "-v", "--capture=sys"])



from probabilit.modeling import Distribution, Constant, Log


def test_garbage_collector():
    loc = Constant(0)
    scale = Constant(1)
    a = Distribution("norm", loc=loc, scale=scale)
    b = Distribution("norm", loc=loc, scale=scale)
    the_sum = a + b
    the_power = the_sum**2
    the_result = Log(1 + the_power)

    # No garbage collection => all nodes have samples_
    the_result.sample(99, random_state=42, gc_strategy=None)
    assert hasattr(loc, "samples_")
    assert hasattr(scale, "samples_")
    assert hasattr(a, "samples_")
    assert hasattr(b, "samples_")
    assert hasattr(the_sum, "samples_")
    assert hasattr(the_power, "samples_")
    assert hasattr(the_result, "samples_")

    # Full garbage collection => only the result has samples_
    the_result.sample(99, random_state=42, gc_strategy=[])
    assert not hasattr(loc, "samples_")
    assert not hasattr(scale, "samples_")
    assert not hasattr(a, "samples_")
    assert not hasattr(b, "samples_")
    assert not hasattr(the_sum, "samples_")
    assert not hasattr(the_power, "samples_")
    assert hasattr(the_result, "samples_")

    # Partial garbage collection => selected nodes have samples_
    the_result.sample(99, random_state=42, gc_strategy=[loc, scale])
    assert hasattr(loc, "samples_")
    assert hasattr(scale, "samples_")
    assert not hasattr(a, "samples_")
    assert not hasattr(b, "samples_")
    assert not hasattr(the_sum, "samples_")
    assert not hasattr(the_power, "samples_")
    assert hasattr(the_result, "samples_")


def test_garbage_collector_large_graph():
    sampling_nodes = []

    result = 0
    for year in range(99):
        addition = Distribution("norm", loc=100, scale=10)
        interest_rate = Distribution("norm", loc=1.05, scale=0.05)
        result = result * interest_rate + addition

        # Store nodes to check them later
        sampling_nodes.append(addition)
        sampling_nodes.append(interest_rate)

    result.sample(99, random_state=42, gc_strategy=[])

    # Verify that GC worked
    assert hasattr(result, "samples_")
    assert not any(hasattr(n, "samples_") for n in sampling_nodes)


if __name__ == "__main__":
    import pytest

    pytest.main(args=[__file__, "-v", "--capture=sys"])
